from gensim.models import Word2Vec
from pathlib import Path
from typing import List, Union
from nltk.tag import pos_tag
import numpy as np


class Word2VecTrainer:  
    """
    Wrapper class for gensim.models.Word2Vec to train a Word2Vec model.

    Methods
    -------
        __init__(model_path, min_count, window, negative, ns_exponent, vector_size, workers, sg, **kwargs)
            Initialize the Word2Vec model
        fit(data, output_path, epochs, start_alpha, end_alpha, compute_loss, **kwargs)
            Train the Word2Vec model on the given data
    """  

    def __init__(
            self,
            **kwargs
            ):
        """
        Args:
            min_count: Ignores all words with total frequency lower than this
            window: The maximum distance between the current and predicted word within a sentence
            negative: If > 0, negative sampling will be used
            ns_exponent: The exponent used to shape the negative sampling distribution
            vector_size: Dimensionality of the word vectors
            workers: Number of worker threads to train the model
            sg: Training algorithm: 1 for skip-gram; otherwise CBOW
            **kwargs: Additional arguments to pass to the gensim.models.Word2Vec constructor
        """
        try:
            self.model = Word2Vec(**kwargs)
        except Exception as e:
            raise RuntimeError(f"Error initializing the Word2Vec model: {e}. Please check the input parameters. Refer to the gensim.models.Word2Vec documentation for more information.")

    @classmethod
    def load(
            cls,
            model_path: Union[str, Path]
            ):
        """
        Load a pretrained Word2Vec model

        Args:
            model_path: Path to the pretrained model

        Returns:
            model: The pretrained Word2Vec model

        Examples:
            >>> from ttta.methods.word2vec.word2vec import Word2VecTrainer
            >>> model = Word2VecTrainer.load('model.model')
            >>> print('Loaded model: ', model)
            Loaded model:  Word2Vec(vocab=5, vector_size=100, alpha=0.025)
        """
        try:
            model = Word2Vec.load(model_path)
            model_args = {
                'vector_size': model.vector_size,
                'workers': model.workers,
                'epochs': model.epochs,
                'batch_words': model.batch_words,
                'sg': model.sg,
                'alpha': model.alpha,
                'min_alpha': model.min_alpha,
                'window': model.window,
                'shrink_windows': model.shrink_windows,
                'hs': model.hs,
                'negative': model.negative,
                'ns_exponent': model.ns_exponent,
                'cbow_mean': model.cbow_mean,
                'compute_loss': model.compute_loss,
                'max_final_vocab': model.max_final_vocab,
                'max_vocab_size': model.max_vocab_size,
                'min_count': model.min_count,
                'sample': model.sample,
                'sorted_vocab': model.sorted_vocab,
                'null_word': model.null_word,
                'seed': model.seed,
            }
            trainer = cls(**model_args)
            trainer.model = model  # Set the loaded model as the trainer's model
            
            return trainer


        except Exception as e:
            raise RuntimeError(f"Error loading the Word2Vec model: {e}. Please check the input path.")

    def train(
            self, 
            data: List[str],
            epochs=5,
            start_alpha=0.025,
            end_alpha=0.0001,
            compute_loss=True,
            update: bool = False, # update=True to update the model
            **kwargs
            ):
        """
        Train the Word2Vec model on the given data
        
        Args:
            data: List of documents
            epochs: Number of epochs
            start_alpha: Learning rate
            end_alpha: Minimum learning rate
            compute_loss: Whether to compute the loss
            **kwargs : optional

        Examples:
            >>> from ttta.methods.word2vec.word2vec import Word2VecTrainer
            >>> texts = ['This is a test.', 'This is another test.', 'This is a third test.']
            >>> Word2VecTrainer().train(texts, epochs=1)
            >>> print('Trained model: ', Word2VecTrainer().model)
            Trained model:  Word2Vec(vocab=5, vector_size=100, alpha=0.025)
        """
        from gensim.models.phrases import Phrases, Phraser
        phrases = Phrases(data, min_count=2, threshold=1.0)
        bigram = Phraser(phrases)  
        sentences = bigram[data]
   
        self.model.build_vocab(sentences, update=update)
  
        if len(self.model.wv.index_to_key) == 0:
            raise RuntimeError("Vocabulary is empty, cannot proceed with training.")
        
        total_examples = self.model.corpus_count
        self.model.train(
                sentences,
                total_examples=total_examples,
                epochs=epochs,
                start_alpha=start_alpha,
                end_alpha=end_alpha,
                compute_loss=compute_loss,
                **kwargs
                )
        self.model.init_sims(replace=True)
        return self.model


    
    def get_model(self) -> Word2Vec:
        """
        Get the trained model

        Returns:
            model: The trained Word2Vec model
        """
        return self.model
    
    def get_vocab(self) -> List[str]:
        """
        Get the vocabulary of the trained model

        Returns:
            vocab: The vocabulary of the model
        """
        return self.model.wv.index_to_key
    
    

class Word2VecAlign:
    """
    Wrapper class for gensim.models.Word2Vec to align Word2Vec models.

    Methods
    -------
        __init__(model_paths)
            Initialize the Word2VecAlign object with a list of paths to the Word2Vec models.
        load_models()
            Load the models
        align_models(reference_index, output_dir, method)
            Align the models
    """
    def __init__(
            self, 
            models: List[Word2Vec]
            ):
        """
        Args:
            models: List of Word2Vec models to align. See gensim.models.Word2Vec for more information.
  
        """
        self.models = models

    def align(
            self,
            reference: int = -1,
            method: str = "procrustes",
            ) -> List[Word2Vec]:
        """
        Align the models

        Args: 
            reference: Index of the reference model
            method: Alignment method, by default "procrustes". Currently, only procrustes is implemented.
      
        Returns:
            aligned_models: List of aligned models

        Examples:
            >>> from from ttta.methods.word2vec.word2vec import Word2VecAlign
            >>> model_paths = ['model1.model', 'model2.model']
            >>> Word2VecAlign(model_paths).align_models(reference_index=0, output_dir='aligned_models')
            >>> print('Aligned models: ', Word2VecAlign(model_paths).aligned_models)
            Aligned models:  [Word2Vec(vocab=5, vector_size=100, alpha=0.025), Word2Vec(vocab=5, vector_size=100, alpha=0.025)]
        """
        
        if method != "procrustes":
            raise NotImplementedError("Only procrustes alignment is implemented. Please use method='procrustes'")

        
        self.reference_model = self.models[reference]
        self.aligned_models: List[Word2Vec] = []

        for i, model in enumerate(self.models):
            if i == reference:
                self.aligned_models.append(model)

            aligned_model = smart_procrustes_align_gensim(self.reference_model,model)
            self.aligned_models.append(aligned_model)
        return self.aligned_models


class Word2VecInference:
    """
    Wrapper class for gensim.models.Word2Vec for Inference.

    Methods
    -------
        __init__(pretrained_model_path)
            Initialize the Word2VecInference object with a pretrained model.
        get_embedding(word, norm)
            Infer the vector of a word
        get_similarity(word1, word2)
            Get the cosine similarity between two words
        get_top_k_words(word, k)
            Get the top k most similar words to a word in the vocabulary of the model.
    """
    def __init__(
            self,
            model: Word2Vec
            ):
        """
        Args:
            model: A Word2Vec model
        """
        self.model = model



    def infer_vector(self, word:str, norm = False) -> List[float]:
        """
        Infer the vector of a word

        Args:
            word: The word to infer the embedding vector of
            norm: Whether to normalize the vector

        Returns:
            embedding: The embedding vector of the word
        """
        return self.model.wv.get_vector(word, norm = norm)
    
    
    def get_similarity(self, word1: str, word2: str) -> float:
        """
        Get the cosine similarity between two words' embedding vectors

        Args:
            word1: The first word
            word2: The second word
        
        Returns:
            similarity: The cosine similarity between the two words
        
        Examples:
            >>> from ttta.methods.word2vec.word2vec import Word2VecInference
            >>> Word2VecInference('model.model').get_similarity('test', 'another')
            0.99999994
        """
        return self.model.wv.similarity(word1, word2)
    
    def get_top_k_words(
            self,
            main_word: str,
            k: int = 10,
            pos_tag: Union[bool, str, List[str]] = False
            ):
        """
        Get the top k most similar words to a word in the vocabulary of the model.

        Args:
            main_word: The word to get the top k most similar words of
            k: The number of words to return
        
        Returns:
            topk: Tuple of lists of the top k most similar words and their cosine similarities
            similarity: The cosine similarities of the top k most similar words
        
        Examples:
            >>> from ttta.methods.word2vec.word2vec import Word2VecInference
            >>> Word2VecInference('model.model').get_top_k_words('test', k=1)
            (['another'], [0.9999999403953552])
        """
        try:
            top_k_words = []
            i = k
            if isinstance(pos_tag, str):
                pos_tag = [pos_tag]

            while len(top_k_words) < k:
                sims = self.model.wv.most_similar(main_word, topn= i)
                words, similarities= tuple(map(list, zip(*sims)))

               
                if isinstance(pos_tag, list):
                    words = [word for word in words if pos_tag([word])[0][1] in pos_tag]
                    
                elif pos_tag:
                    main_pos = pos_tag([main_word])[0][1]
                    words = [word for word in words if pos_tag([word])[0][1] == main_pos]

                top_k_words.extend(words)
                i += 1

            return top_k_words, similarities
        
        except KeyError:
            raise ValueError(f"Word '{main_word}' not in vocabulary. Please try another word.")



# ------------------- smart_procrustes_align_gensim --------------------
# from: https://gist.github.com/quadrismegistus/09a93e219a6ffc4f216fb85235535faf
def smart_procrustes_align_gensim(base_embed, other_embed, words=None):
    """
    Original script: https://gist.github.com/quadrismegistus/09a93e219a6ffc4f216fb85235535faf
    Procrustes align two gensim word2vec models (to allow for comparison between same word across models).
    Code ported from HistWords <https://github.com/williamleif/histwords> by William Hamilton <wleif@stanford.edu>.
        
    First, intersect the vocabularies (see `intersection_align_gensim` documentation).
    Then do the alignment on the other_embed model.
    Replace the other_embed model's syn0 and syn0norm numpy matrices with the aligned version.
    Return other_embed.

    If `words` is set, intersect the two models' vocabulary with the vocabulary in words (see `intersection_align_gensim` documentation).
    """

    # patch by Richard So [https://twitter.com/richardjeanso) (thanks!) to update this code for new version of gensim
    # base_embed.init_sims(replace=True)
    # other_embed.init_sims(replace=True)

    # make sure vocabulary and indices are aligned
    in_base_embed, in_other_embed = intersection_align_gensim(base_embed, other_embed, words=words)

    # re-filling the normed vectors
    in_base_embed.wv.fill_norms(force=True)
    in_other_embed.wv.fill_norms(force=True)

    # get the (normalized) embedding matrices
    base_vecs = in_base_embed.wv.get_normed_vectors()
    other_vecs = in_other_embed.wv.get_normed_vectors()

    # just a matrix dot product with numpy
    m = other_vecs.T.dot(base_vecs) 
    # SVD method from numpy
    u, _, v = np.linalg.svd(m)
    # another matrix operation
    ortho = u.dot(v) 
    # Replace original array with modified one, i.e. multiplying the embedding matrix by "ortho"
    other_embed.wv.vectors = (other_embed.wv.vectors).dot(ortho)    
    
    return other_embed

# ------------------- intersection_align_gensim --------------------
# from: https://gist.github.com/quadrismegistus/09a93e219a6ffc4f216fb85235535faf
def intersection_align_gensim(m1, m2, words=None):
    """
    Intersect two gensim word2vec models, m1 and m2.
    Only the shared vocabulary between them is kept.
    If 'words' is set (as list or set), then the vocabulary is intersected with this list as well.
    Indices are re-organized from 0..N in order of descending frequency (=sum of counts from both m1 and m2).
    These indices correspond to the new syn0 and syn0norm objects in both gensim models:
        -- so that Row 0 of m1.syn0 will be for the same word as Row 0 of m2.syn0
        -- you can find the index of any word on the .index2word list: model.index2word.index(word) => 2
    The .vocab dictionary is also updated for each model, preserving the count but updating the index.
    """

    # Get the vocab for each model
    vocab_m1 = set(m1.wv.index_to_key)
    vocab_m2 = set(m2.wv.index_to_key)

    # Find the common vocabulary
    common_vocab = vocab_m1 & vocab_m2
    if words: common_vocab &= set(words)

    # If no alignment necessary because vocab is identical...
    if not vocab_m1 - common_vocab and not vocab_m2 - common_vocab:
        return (m1,m2)

    # Otherwise sort by frequency (summed for both)
    common_vocab = list(common_vocab)
    common_vocab.sort(key=lambda w: m1.wv.get_vecattr(w, "count") + m2.wv.get_vecattr(w, "count"), reverse=True)

    # Then for each model...
    for m in [m1, m2]:
        # Replace old syn0norm array with new one (with common vocab)
        indices = [m.wv.key_to_index[w] for w in common_vocab]
        old_arr = m.wv.vectors
        new_arr = np.array([old_arr[index] for index in indices])
        m.wv.vectors = new_arr

        # Replace old vocab dictionary with new one (with common vocab)
        # and old index2word with new one
        new_key_to_index = {}
        new_index_to_key = []
        for new_index, key in enumerate(common_vocab):
            new_key_to_index[key] = new_index
            new_index_to_key.append(key)
        m.wv.key_to_index = new_key_to_index
        m.wv.index_to_key = new_index_to_key
        
    return m1, m2



import os
import pickle
import matplotlib
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from scipy.spatial.distance import cosine
from typing import Union, List, Tuple, Callable
from tqdm import tqdm
from .lda_prototype import LDAPrototype
from datetime import datetime
import warnings
import seaborn as sns
from operator import itemgetter
import math
from ..preprocessing import chunk_creation

class RollingLDA:
    def __init__(self, K: int, how: Union[str, List[datetime]] = "M", warmup: int = 48, memory: int = 3, alpha: float = None, gamma: float = None,
                 initial_epochs: int = 100, subsequent_epochs: int = 50, min_count: int = 2, max_assign=False, prototype: int = 10,
                 topic_threshold: List[Union[int, float]] = None, prototype_measure: Union[str, Callable] = "jaccard", lda: LDAPrototype = None,
                 min_docs_per_chunk: int = None, verbose: int = 1) -> None:
        """
            Initializes a RollingLDA model.
            Args:
                K: The number of topics.
                how: List of datetime dates indicating the end of time chunks or a string indicating the frequency of the time chunks as in pandas.resample().
                warmup: The number of chunks to use for the initial fitting period of RollinglDA.
                memory: The number of chunks to look back for when calculating the topic proportions for the RollingLDA.
                alpha: The alpha parameter of the LDA model.
                gamma: The gamma parameter of the LDA model.
                initial_epochs: The number of epochs to train the LDA model in the initial fit.
                subsequent_epochs: The number of epochs to train the LDA model in the subsequent fits.
                min_count: The minimum number of times a word must occur in the corpus to be included in the vocabulary.
                max_assign: Whether the maximum number of topic assignments per document should be used as the final topic assignment.
                prototype: Number of LDAs to fit and to choose a Prototype from.
                topic_threshold: The minimum occurrences within a topic to be relevant for the prototype similarity calculation.
                prototype_measure: The measure to use for calculating the distance between the LDAs. Can be "jaccard", "cosine" or a custom function.
                lda: An LDAPrototype model to use instead of fitting a new one.
                min_docs_per_chunk: The minimum number of documents a chunk must contain to be used for the LDA training.
                verbose: The verbosity of the output. 0 does not show any output, 1 shows a progress bar, 2 also shows information relevant for debugging.
        """
        if not isinstance(how, str) and not isinstance(how, list):
            raise TypeError("how must be a string or a list of datetime dates!")
        if not isinstance(warmup, int):
            try:
                if warmup == int(warmup):
                    warmup = int(warmup)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("warmup must be an integer!")
        if warmup < 0:
            raise ValueError("warmup must be a natural number")
        if not isinstance(memory, int):
            try:
                if memory == int(memory):
                    memory = int(memory)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("memory must be an integer!")
        if not isinstance(K, int):
            try:
                if K == int(K):
                    K = int(K)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("K must be an integer!")
        if memory < 1:
            raise ValueError("memory must be a natural number greater than 0")
        if not isinstance(initial_epochs, int):
            try:
                if initial_epochs == int(initial_epochs):
                    initial_epochs = int(initial_epochs)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("initial_epochs must be an integer!")
        if not isinstance(subsequent_epochs, int):
            try:
                if subsequent_epochs == int(subsequent_epochs):
                    subsequent_epochs = int(subsequent_epochs)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("subsequent_epochs must be an integer!")
        if initial_epochs < 1 or subsequent_epochs < 1:
            raise ValueError("initial_epochs and subsequent_epochs must be natural numbers greater than 0")
        if min_docs_per_chunk is None:
            min_docs_per_chunk = K * 10
        if not isinstance(min_docs_per_chunk, int):
            try:
                if min_docs_per_chunk == int(min_docs_per_chunk):
                    min_docs_per_chunk = int(min_docs_per_chunk)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("min_docs_per_chunk must be an integer or None!")
        if min_docs_per_chunk < 1:
            raise ValueError("min_docs_per_chunk must be a natural number greater than 0")
        self._min_docs_per_chunk = min_docs_per_chunk
        if not topic_threshold:
            topic_threshold = [5, 0.002]
        self._K = K
        self._how = how
        self._warmup = warmup
        self._memory = memory
        if alpha is None:
            self._alpha = 1 / self._K
        else:
            self._alpha = alpha
        if gamma is None:
            self._gamma = 1 / self._K
        else:
            self._gamma = gamma
        self._initial_epochs = initial_epochs
        self._subsequent_epochs = subsequent_epochs
        self._prototype = prototype
        self._threshold = topic_threshold
        self._measure = prototype_measure
        self._min_count = min_count
        self._max_assign = max_assign
        self.lda = lda if lda is not None else LDAPrototype(K=K, alpha=alpha, gamma=gamma, prototype=prototype, topic_threshold=topic_threshold,
                                                            prototype_measure=prototype_measure, min_count=min_count, max_assign=max_assign,
                                                            verbose=verbose-1)
        self._updated_how = []
        self._verbose = verbose
        self._distances_simulated = None
        self._distances_observed = None
        self.chunk_indices = None
        self._date_column = "date"
        self._text_column = "text"
        self._last_text = None

    def fit(self, texts: pd.DataFrame, workers: int = 1, text_column: str = "text", date_column: str = "date") -> None:
        """
        Fits a RollingLDA model in place to the given texts from scratch. When updating an existing model, use fit_update() instead.
        Args:
            texts: A pandas DataFrame containing the columns text_column and date_column containing the documents and their respective dates.
                   The dates must be in a format interpretable by pandas.to_datetime(). Each element of the text_column column must be a list of strings.
            workers: The number of workers to use for parallelization.
            text_column: The name of the column in texts containing the documents.
            date_column: The name of the column in texts containing the dates.
        Returns:
            None
        """
        if not isinstance(texts, pd.DataFrame):
            raise TypeError("texts must be a pandas DataFrame!")
        if not isinstance(text_column, str):
            raise TypeError("text_column must be a string!")
        if text_column not in texts.columns:
            raise ValueError("texts must contain the column specified in text_column!")
        if not isinstance(date_column, str):
            raise TypeError("date_column must be a string!")
        if date_column not in texts.columns:
            raise ValueError("texts must contain the column specified in date_column!")
        self._text_column = text_column
        self._date_column = date_column
        if not isinstance(texts[self._text_column].iloc[0], list):
            raise TypeError("The elements of the 'texts' column of texts must each contain a tokenized document as a list of strings!")
        if not isinstance(workers, int):
            try:
                if workers == int(workers):
                    workers = int(workers)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("workers must be an integer!")
        if self.lda.is_trained():
            raise ValueError("The LDA has already been trained. If you want to update the training process, use fit_update() instead!")

        texts[date_column] = pd.to_datetime(texts[date_column])
        texts.sort_values(by=date_column, inplace=True)
        self.chunk_indices = self._get_time_indices(texts)
        if self.chunk_indices.shape[0] < self._warmup + 1:
            raise ValueError(f"The number of chunks in the data must be larger than warmup!")
        iterator = self.chunk_indices.iloc[self._warmup:].iterrows()
        if self._verbose > 0:
            iterator = tqdm(iterator, unit="chunk")
        for i, row in iterator:
            if self._verbose and i < len(self.chunk_indices) - 1:
                iterator.set_description(f"Processing {int(self.chunk_indices.iloc[i + 1]['chunk_start'] - 1 - self.chunk_indices.iloc[i]['chunk_start'])} documents in "
                                         f"chunk {self.chunk_indices.iloc[i][date_column].strftime('%Y-%m-%d')}")
            if i == self._warmup:  # fit warmup chunks
                self.lda.fit(texts[text_column], epochs=self._initial_epochs,
                             chunk_end=self.chunk_indices.iloc[self._warmup + 1]["chunk_start"], memory_start=0, workers=workers)
                continue
            end = len(texts) if i + 1 >= len(self.chunk_indices) else self.chunk_indices.iloc[i + 1]["chunk_start"] - 1
            self.lda.fit(texts[text_column], epochs=self._subsequent_epochs, first_chunk=False, chunk_end=end, memory_start=row["memory_start"], workers=workers)
        self.chunk_indices["chunk_start_preprocessed"] = [sum([x < y for x in self.lda._deleted_indices]) for y in self.chunk_indices["chunk_start"]]
        if self._last_text is None:
            self._last_text = {date_column: None, "index": 0}
        self._last_text[date_column] = texts[date_column].iloc[-1]
        self._last_text["index"] += len(texts) - 1
        print(self._last_text)

    def fit_update(self, texts: pd.DataFrame, how: Union[str, List[datetime]] = None, workers: int = 1, text_column: str = "text",
                   date_column: str = "date") -> None:
        """
        Updates the fit of a RollingLDA model in place to the given texts from scratch. When training a model from scratch, fit() is called instead.
        Args:
            texts: A pandas DataFrame containing the columns text_column and date_column containing the documents and their respective dates.
                   The dates must be in a format interpretable by pandas.to_datetime(). Each element of the text_column column must be a list of strings.
            how: List of datetime dates indicating the end of time chunks or a string indicating the frequency of the time chunks as in pandas.resample().
            workers: The number of workers to use for parallelization.
            text_column: The name of the column in texts containing the documents.
            date_column: The name of the column in texts containing the dates.
        Returns:
            None
        """
        if not isinstance(texts, pd.DataFrame):
            raise TypeError("texts must be a pandas DataFrame!")
        if not isinstance(text_column, str):
            raise TypeError("text_column must be a string!")
        if text_column not in texts.columns:
            raise ValueError("texts must contain the column specified in text_column!")
        if text_column != self._text_column:
            texts[text_column] = texts[self._text_column]
        if not isinstance(date_column, str):
            raise TypeError("date_column must be a string!")
        if date_column not in texts.columns:
            raise ValueError("texts must contain the column specified in date_column!")
        if date_column != self._date_column:
            texts[date_column] = texts[self._date_column]
        if not isinstance(texts[self._text_column].iloc[0], list):
            raise TypeError("The elements of the 'texts' column of texts must each contain a tokenized document as a list of strings!")
        if not isinstance(how, str) and not isinstance(how, list) and how is not None:
            raise TypeError("how must be a string or a list of datetime dates!")
        if not isinstance(workers, int):
            try:
                if workers == int(workers):
                    workers = int(workers)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("workers must be an integer!")
        if not self.lda.is_trained():
            warnings.warn("The LDA has not been trained yet. Will use fit() instead.")
            self.fit(texts, workers)
            return
        last_trained = len(self.chunk_indices)
        texts[self._date_column] = pd.to_datetime(texts[self._date_column])
        texts.sort_values(by=self._date_column, inplace=True)
        self.chunk_indices = self._get_time_indices(texts, update=True, how=how)
        iterator = self.chunk_indices.iloc[last_trained:].iterrows()
        if self._verbose > 0:
            iterator = tqdm(iterator, unit="chunk")
        for i, row in iterator:
            if self._verbose and i < len(self.chunk_indices) - 1:
                iterator.set_description(f"Processing {self.chunk_indices.iloc[i + 1]['chunk_start'] - 1 - self.chunk_indices.iloc[i]['chunk_start']} documents in "
                                         f"chunk {self.chunk_indices.iloc[i]['date'].strftime('%Y-%m-%d')}")
            end = self.chunk_indices.iloc[i + 1]["chunk_start"] if i + 1 < len(self.chunk_indices) else len(texts) + self._last_text["index"] + 1   # todo hier ueberpruefen, ob ohne self.text richtige indices berechnet werden
            self.chunk_indices["chunk_start_preprocessed"] = [sum([x < y for x in self.lda._deleted_indices]) for y in self.chunk_indices["chunk_start"]]
            self.lda._last_end[-1] = self.chunk_indices.iloc[i]["chunk_start_preprocessed"] - self.chunk_indices.iloc[last_trained]["chunk_start_preprocessed"]
            self.lda.fit(texts[self._text_column], epochs=self._subsequent_epochs, first_chunk=False, chunk_end=end - self._last_text["index"] - 1, memory_start=row["memory_start"], workers=workers)
        self._last_text[self._date_column] = texts[self._date_column].iloc[-1]
        self._last_text["index"] += len(texts)
        self.chunk_indices["chunk_start_preprocessed"] = [sum([x < y for x in self.lda._deleted_indices]) for y in self.chunk_indices["chunk_start"]]

    def top_words(self, chunk: Union[int, str] = None, topic: int = None, number: int = 5, importance: bool = True,
                  return_as_data_frame: bool = True) -> Union[List[str], List[List[str]]]:
        """
            Returns the top words for the given chunk and topic.
            Args:
                chunk: The chunk for which the top words should be returned. If None, the top words over the entire time frame are returned.
                       If "all", the top words for each individual chunk are returned.
                topic: The topic for which the top words should be returned. If None, the top words for all topics are returned.
                number: The number of top words to return.
                importance: Whether the words should be weighted based on their importance to a topic or their absolute frequency.
                return_as_data_frame: Whether the top words should be returned as a pandas DataFrame.
        """
        if not isinstance(chunk, int) and chunk is not None and chunk != "all":
            try:
                if chunk == int(chunk):
                    chunk = int(chunk)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("chunk must be an integer, 'all' or None!")
        if not isinstance(topic, int) and topic is not None:
            try:
                if topic == int(topic):
                    topic = int(topic)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("topic must be an integer or None!")
        if not isinstance(number, int):
            try:
                if number == int(number):
                    number = int(number)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("number must be an integer!")
        if not isinstance(importance, bool):
            raise TypeError("importance must be a boolean!")
        if chunk is None:
            word_topic_matrix = self.get_word_topic_matrix()
            return self.lda.top_words(number=number, topic=topic, importance=importance, word_topic_matrix=word_topic_matrix,
                                      return_as_data_frame=return_as_data_frame)
        elif chunk == "all":
            top_words = [self.top_words(chunk=i, number=number, importance=importance, return_as_data_frame=return_as_data_frame) for i in range(len(self.chunk_indices))]
            if return_as_data_frame:
                top_words = pd.concat(top_words)
                top_words.index = [f"Chunk {i+1}, word {x+1}" for i in range(len(self.chunk_indices)) for x in range(number)]
            return top_words
        else:
            word_topic_matrix = self.get_word_topic_matrix(chunk)
            return self.lda.top_words(number=number, topic=topic, importance=importance, word_topic_matrix=word_topic_matrix,
                                      return_as_data_frame=return_as_data_frame)

    def get_word_topic_matrix(self, chunk: int = None) -> np.ndarray:
        """
            Returns the topic assignments for the given chunk. If no chunk is given, the topic assignments for all chunks are returned.
            Args:
                chunk: The chunk for which the topic assignments should be returned. If None, the topic assignments for all chunks are returned.
            Returns:
                The topic assignments for the given chunk or all chunks.
        """
        if not isinstance(chunk, int) and chunk is not None:
            try:
                if chunk == int(chunk):
                    chunk = int(chunk)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("chunk must be an integer or None!")
        assignments = self.lda.get_assignment_vec()
        words = self.lda.get_word_vec()
        len_of_docs = self.lda._len_of_docs
        if chunk is None:
            return self.lda.get_word_topic_matrix(words, assignments)
        else:
            start_index = len_of_docs[:self.chunk_indices.iloc[chunk]["chunk_start_preprocessed"]].sum()
            if chunk == len(self.chunk_indices) - 1:
                end_index = len(assignments)
            else:
                end_index = len_of_docs[:self.chunk_indices.iloc[chunk + 1]["chunk_start_preprocessed"]].sum()
            return self.lda.get_word_topic_matrix(words[start_index:end_index], assignments[start_index:end_index])

    def get_parameters(self) -> dict:
        """
            Returns the parameters of the RollingLDA model.
            Returns:
                A dictionary containing the parameters of the RollingLDA model.
        """
        return self.__dict__.copy()

    def set_parameters(self, parameters: dict) -> None:
        """
            Sets the parameters of the RollingLDA model.
            Args:
                parameters: A dictionary containing the parameters of the RollingLDA model.
            Returns:
                None
        """
        if not isinstance(parameters, dict):
            raise TypeError("parameters must be a dictionary!")
        self.__dict__.update(parameters)

    def save(self, path: str) -> None:
        """
            Saves the RollingLDA model to the given path as a .pickle-file.
            Args:
                path: The path to which the RollingLDA model should be saved.
            Returns:
                None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        pickle.dump(self, open(path, "wb"))

    def load(self, path: str) -> None:
        """
            Loads a pickled RollingLDA model from the given path.
            Args:
                path: The path from which the RollingLDA model should be loaded.
            Returns:
                None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        loaded = pickle.load(open(path, "rb"))
        self.__dict__ = loaded.__dict__

    def get_document_topic_matrix(self, chunk: int = None) -> np.ndarray:
        """
            Returns the document-topic matrix for the given chunk. If no chunk is given, the document-topic matrix for all chunks is returned.
            Args:
                chunk: The chunk for which the document-topic matrix should be returned. If None, the document-topic matrix for all chunks is returned.
            Returns:
                The document-topic matrix for the given chunk or all chunks.
        """
        if not isinstance(chunk, int) and chunk is not None:
            try:
                if chunk == int(chunk):
                    chunk = int(chunk)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("chunk must be an integer or None!")
        if chunk is None:
            all_matrices = []
            for i in range(len(self.chunk_indices)):
                all_matrices.append(self.get_document_topic_matrix(i))
            return np.concatenate(all_matrices, axis=0)
        else:
            assignments = self.lda.get_assignment_vec()
            docs = self.lda.get_doc_vec()
            len_of_docs = self.lda._len_of_docs
            start_index = len_of_docs[:self.chunk_indices.iloc[chunk]["chunk_start_preprocessed"]].sum()
            if chunk == len(self.chunk_indices) - 1:
                end_index = len(assignments)
            else:
                end_index = len_of_docs[:self.chunk_indices.iloc[chunk + 1]["chunk_start_preprocessed"]].sum()
            return self.lda.get_document_topic_matrix(docs[start_index:end_index],
                                                      assignments[start_index:end_index])


    def _get_time_indices(self, texts: pd.DataFrame, update: bool = False, how: Union[str, List[datetime]] = None) -> pd.DataFrame:
        """
            Creates the time indices for the given texts. If update is True, the time indices are appended to the existing ones.
            Args:
                texts: A pandas DataFrame containing the documents and their respective dates.
                       The dates must be in a format interpretable by pandas.to_datetime(). The texts must be a list of strings.
                update: Whether the time indices should be appended to the existing ones.
                how: List of datetime dates indicating the end of time chunks or a string indicating the frequency of the time chunks. Used to create time
                     chunks when fixed dates were used for the initial fit. If None, the same time chunk rule as in the initial fit is used.
        """
        if not isinstance(texts, pd.DataFrame):
            raise TypeError("texts must be a pandas DataFrame!")
        if not isinstance(update, bool):
            raise TypeError("update must be a boolean!")
        if how is None:
            how = self._how
        elif isinstance(how, str):
            if isinstance(self._how, list):
                warnings.warn(f"The time indices are created using periodic distances instead of fixed dates. This might create inconsistencies.")
            elif self._how != how:
                warnings.warn(f"The time indices are created using {how} instead of {self._how}. This might create inconsistencies.")
        else:
            if isinstance(self._how, str):
                warnings.warn(f"The time indices are created using fixed dates instead of periodic distances. This might create inconsistencies.")
        last_date = self._last_text[self._date_column] if self.lda.is_trained() else None
        period_start = chunk_creation._get_time_indices(texts, how, last_date=last_date, date_column=self._date_column,
                                                        min_docs_per_chunk=self._min_docs_per_chunk)
        if update:
            period_start["chunk_start"] += self._last_text["index"] + 1
            memory_start = [period_start["chunk_start"].iloc[i - self._memory] if i - self._memory >= 0 else self.chunk_indices["chunk_start"].iloc[i - self._memory] for i in range(len(period_start))]
        else:
            memory_start = [0 if i <= self._warmup else period_start["chunk_start"].iloc[max(i - self._memory, 0)] for i in range(len(period_start))]
        period_start["memory_start"] = memory_start
        period_start = self.chunk_indices._append(period_start) if update else period_start
        period_start = period_start.reset_index(drop=True)

        if isinstance(how, list) and isinstance(self._how, list):
            self._how.append(how)
        elif isinstance(how, str) and isinstance(self._how, str):
            if how != self._how:
                self._updated_how.append(how)
        else:
            self._updated_how.append(how)
        return period_start


    def wordclouds(self, chunks: List[int] = None, topic: int = None, number: int = 50, path: str = "wordclouds",
                   height: int = 500, width: int = 700, show: bool = True):
        """
            Plots the wordclouds for the given chunk. If chunk is None, the wordclouds are plotted for every time chunk.
            Args:
                chunks: The chunks for which the wordclouds should be plotted. If None, the wordclouds over the entire time frame are plotted.
                topic: The topic to create a word cloud for or None for all topics
                number: number of top words to include in the word cloud
                path: path to a directory to store the wordcloud files in
                width: width of the image
                height: height of the image
                show: should the image be shown
            Returns:
                None
        """
        if not isinstance(chunks, List) and chunks is not None:
            raise TypeError("chunks must be a list or None!")
        if chunks and not all([isinstance(x, int) for x in chunks]):
            raise TypeError("chunks must be a list of integers!")
        if chunks is None:
            chunks = range(len(self.chunk_indices))
        for chunk in chunks:
            word_topic_matrix = self.get_word_topic_matrix(chunk)
            if path is not None and not os.path.exists(path):
                os.makedirs(path)
                path += "/" if path[-1] != "/" else ""
            self.lda.wordclouds(topic=topic, number=number, path=os.path.join(path, f"chunk_{chunk}.pdf"), height=height, width=width,
                                show=show, word_topic_matrix=word_topic_matrix)




import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from typing import List
from tqdm import tqdm
from transformers import BertForMaskedLM, AdamW, AutoTokenizer
import logging

class NewsExamplesDataset(Dataset):
    """
    Dataset generator for the news examples

    Attributes
    ----------
        examples: List[str]
            List of the examples
        words_to_mask: List[str]
            List of the words to mask

    Methods
    -------
        _pick_masked_embeddings(self, encodings: Dict[str, List[int]]) -> List[int]
            Returns the masked embeddings
        __len__(self) -> int
            Returns the length of the dataset
        __getitem__(self, idx: int) -> Dict[str, torch.Tensor]
            Returns the item at the given index

    """
    def __init__(self,
                 examples:List[str],
                 words_to_mask: List[str],
                 chunk_size=419,

        ):

        print(f'{"-" * 10} Initializing the dataset {"-" * 10}')
        self.examples = examples
        self.tokenizer = AutoTokenizer.from_pretrained('bert-base-uncased')
        self.encodings = self.tokenizer(self.examples, return_tensors='pt', max_length=chunk_size, truncation=True, padding="max_length")
        self.encodings["word_ids"] = [self.encodings.word_ids(i) for i in range(len(self.encodings["input_ids"]))]
        self.encodings["labels"] = self.encodings["input_ids"].detach().clone()
        self.words = self.tokenizer.convert_tokens_to_ids(words_to_mask)

        self.masked_embeddings = self._pick_masked_embeddings(self.encodings['input_ids'])
        self.max_length = chunk_size

    def _pick_masked_embeddings(self, inputs:torch.Tensor, percentage_size=0.15):
        """
        Returns the list of embeddings that will be masked
        Args:
            inputs: torch.Tensor
                The input tensor
            percentage_size: float
                The percentage of the embeddings to mask
        Returns:
            List[torch.Tensor]
        """
        dim = inputs.shape[0]
        idx = np.random.choice(list(range(dim)), int(dim * percentage_size), replace=False)
        docs = inputs[idx]

        for idx, doc in enumerate(docs):
            doc_ids = doc.numpy().flatten()
            doc_ids[(doc_ids == self.words[0]) | (doc_ids == self.words[1]) | (doc_ids == self.words[2])] = 103
            docs[idx] = torch.tensor(doc_ids)

        return docs
    def __getitem__(self, idx):
        return dict(
            input_ids = self.encodings['input_ids'][idx].clone().detach(),
            attention_mask = self.encodings['attention_mask'][idx].detach().clone(),
            labels = self.encodings['labels'][idx].detach().clone(),
        ).copy()
    def __len__(self):
        return self.encodings.input_ids.shape[0]

class BertTrainor():
    """
    Class for training the Bert model

    Attributes
    ----------
        train_dataset: NewsExamplesDataset
            The dataset to train on
        device: str
            The device to train on
        epochs: int
            The number of epochs to train for

    Methods
    -------
        train(self)
            Trains the model
    """
    def __init__(
            self,
            train_dataset,
            device='cpu',
            epochs=1,
        ):
        self.train_dataset = train_dataset
        self.device = device
        self.epochs = epochs

        self.model = BertForMaskedLM.from_pretrained('bert-base-uncased').to(self.device)
        self.optimizer = AdamW(self.model.parameters(), lr=5e-5)
        self.model.train()
        self.train_dataloader = DataLoader(
            self.train_dataset,
            batch_size=16,
            shuffle=True,
        )

    def train(self):
        for epoch in range(self.epochs):
            loop = tqdm(self.train_dataloader, leave=True)
            for batch in loop:
                self.optimizer.zero_grad()

                input_ids = batch['input_ids'].to(self.device)
                attention_mask = batch['attention_mask'].to(self.device)
                labels = batch['labels'].to(self.device)

                # process
                outputs = self.model(input_ids, attention_mask=attention_mask, labels=labels)

                loss = outputs.loss
                loss.backward()
                self.optimizer.step()

                # print relevant info to progress bar
                loop.set_description(f'Epoch {epoch}')
                loop.set_postfix(loss=loss.item())

        self.model.save_pretrained("bert_model_new")

if __name__ == '__main__':

    with open('../data/all_sentences.txt', 'r') as f:
        sentences = f.read()

    samples = sentences.split('\n')
    data = NewsExamplesDataset(samples, chunk_size=420, words_to_mask=["abuse", "fight", "market"])
    BertTrainor(data, device='cpu', epochs=2).train()




import os
import shutil
import unittest
import numpy as np
import pandas as pd
import sys
sys.path.append('../src')
from ttta.methods.rolling_lda import RollingLDA
from ttta.methods.topical_changes import TopicalChanges

class TestTopicalChanges(unittest.TestCase):
    def setUp(self):
        self.texts = [["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"]]
        self.dates = [pd.to_datetime("2021-01-01"), pd.to_datetime("2021-01-02"), pd.to_datetime("2021-01-03")] * 7
        self.df = pd.DataFrame({"text": self.texts, "date": self.dates})
        self.df.sort_values("date", inplace=True)
        self.roll = RollingLDA(2.0, prototype=3.0, how="1D", warmup=1.0, min_docs_per_chunk=1.0, verbose=0.0,
                          topic_threshold=[1.0, 0.0], initial_epochs=20.0, subsequent_epochs=10.0,
                          memory=1.0, min_count=1.0)
        self.roll.fit(self.df)

    def test_init(self):
        change_obj = TopicalChanges(self.roll, mixture="0.9", reference_period=4.0, quantile_threshold="0.8", samples=5.0)
        self.assertIsInstance(change_obj, TopicalChanges)
        self.assertEqual(change_obj._roll, self.roll)
        self.assertEqual(change_obj.mixture, 0.9)
        self.assertEqual(change_obj.reference_period, 4)
        self.assertEqual(change_obj.quantile_threshold, 0.8)
        self.assertEqual(change_obj.samples, 5)

    def test_init_invalid(self):
        with self.assertRaises(TypeError):
            TopicalChanges(1)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, mixture="a")
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, mixture=1.1)
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, mixture=-1.1)
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, reference_period=0)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, reference_period=1.5)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, reference_period="a")
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, quantile_threshold="a")
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, quantile_threshold=-0.5)
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, quantile_threshold=1.5)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, samples="a")
        with self.assertRaises(ValueError):
            TopicalChanges(self.roll, samples=0)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, samples=1.5)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, save_path=5)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, load_path=5)
        with self.assertRaises(TypeError):
            TopicalChanges(self.roll, fast=5)

    def test_save_and_load(self):
        change_obj1 = TopicalChanges(self.roll, save_path="topical_changes_test", samples=5)
        change_obj2 = TopicalChanges(self.roll, load_path="topical_changes_test", samples=5)
        self.assertEqual(change_obj1._roll, change_obj2._roll)
        self.assertEqual(change_obj1.mixture, change_obj2.mixture)
        self.assertEqual(change_obj1.reference_period, change_obj2.reference_period)
        self.assertEqual(change_obj1.quantile_threshold, change_obj2.quantile_threshold)
        self.assertEqual(change_obj1.samples, change_obj2.samples)
        self.assertTrue(np.allclose(change_obj1._distances_observed, change_obj2._distances_observed))
        self.assertTrue(np.allclose(change_obj1._distances_simulated, change_obj2._distances_simulated))
        shutil.rmtree("topical_changes_test")

    def test_sample_invalid(self):
        change_obj = TopicalChanges(self.roll, samples=50)
        with self.assertRaises(TypeError):
            change_obj._sample(1, 1, np.array([1] * 2))
        with self.assertRaises(TypeError):
            change_obj._sample(np.array([0.5] * len(self.roll.lda.get_vocab())), 1.5, np.array([1] * 2))
        with self.assertRaises(ValueError):
            change_obj._sample(np.array([0.5] * len(self.roll.lda.get_vocab())), -1, np.array([1] * 2))
        with self.assertRaises(TypeError):
            change_obj._sample(np.array([0.5] * len(self.roll.lda.get_vocab())), 1, 1)

    def test_word_impact(self):
        change_obj = TopicalChanges(self.roll, samples=50)
        change_obj.word_impact(number=5.0)
        change_obj.word_impact(number=2, fast=False)
        with self.assertRaises(TypeError):
            change_obj.word_impact(number="a")
        with self.assertRaises(TypeError):
            change_obj.word_impact(number=5.5)
        with self.assertRaises(ValueError):
            change_obj.word_impact(number=-5)
        with self.assertRaises(TypeError):
            change_obj.word_impact(date_format=1)
        with self.assertRaises(TypeError):
            change_obj.word_impact(fast="a")

    def test_plot_distances(self):
        change_obj = TopicalChanges(self.roll, samples=50)
        change_obj.plot_distances()



import os.path
import torch
from pathlib import Path
from transformers import BertTokenizer, BertModel
from transformers import logging as lg
from src.ttta.preprocessing.schemas import WordSenseEmbedding
from typing import Union, List
import nltk
import logging
from nltk.corpus import wordnet as wn
from nltk.corpus import brown
from tqdm import tqdm
from wasabi import msg

logger = logging.getLogger(__name__)
logging.basicConfig(level='INFO')

for corpus in ['wordnet', 'omw-1.4', 'brown']:
    try:
        nltk.find(f'corpora/{corpus}')
    except LookupError:
        logging.info(f'Downloading {corpus} corpus')
        nltk.download(corpus)


class VectorEmbeddings:
    """
    This class is used to infer the vector embeddings of a word from a sentence.

    Methods
    -------
        infer_vector(doc:str, main_word:str)
            This method is used to infer the vector embeddings of a word from a sentence.
        _bert_case_preparation()
            This method is used to prepare the BERT model for the inference.
    """
    model: BertModel
    bert_tokenizer: BertTokenizer
    tokens_tensor: torch.Tensor
    segments_tensors: torch.Tensor
    _tokens: List

    def __init__(
            self,
            pretrained_model_path: Union[str, Path] = None,
    ):
        """
        Initialize the class with the pretrained model path.
        Args:
            pretrained_model_path: Path to the pretrained model.
        """
        self.model_path = pretrained_model_path
        if pretrained_model_path is not None:
            if not os.path.exists(pretrained_model_path):
                raise ValueError(
                    f'The path {pretrained_model_path} does not exist'
                )
            self.model_path = Path(pretrained_model_path)

        self._tokens = []
        self.model = None
        self.vocab = False

        lg.set_verbosity_error()
        self._bert_case_preparation()

    @staticmethod
    def load(pretrained_model_path: Union[str, Path] = None) -> 'VectorEmbeddings':
        """
        Loads the pretrained model from the path provided.
        Args:
            pretrained_model_path: Path to the pretrained model.

        Returns:
            VectorEmbeddings
        """
        if pretrained_model_path is not None:
            if not isinstance(pretrained_model_path, str) or not isinstance(pretrained_model_path, Path):
                raise ValueError(
                    f'The path provided is not a valid path: {pretrained_model_path}'
                )
            return VectorEmbeddings(pretrained_model_path)
        return VectorEmbeddings()

    @property
    def tokens(self):
        return self._tokens

    def _bert_case_preparation(self) -> None:
        """
        This method is used to prepare the BERT model for the inference.
        """
        model_path = self.model_path if self.model_path is not None else 'bert-base-uncased'
        self.bert_tokenizer = BertTokenizer.from_pretrained(model_path)
        self.model = BertModel.from_pretrained(
            model_path,
            output_hidden_states=True,
        )
        self.model.eval()
        self.vocab = True

    def infer_vector(self, doc: str, main_word: str) -> torch.Tensor:
        """
        This method is used to infer the vector embeddings of a word from a sentence.
        Args:
            doc: Document to process
            main_word: Main work to extract the vector embeddings for.

        Returns: torch.Tensor

        """
        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.model.__class__.__name__} has not been initialized'
            )
        if main_word not in doc:
            raise ValueError(
                f'The word: "{main_word}" does not exist in the document: {doc}'
            )

        marked_text = "[CLS] " + doc + " [SEP]"
        tokens = self.bert_tokenizer.tokenize(marked_text)
        try:
            main_token_id = tokens.index(main_word.lower())
            idx = self.bert_tokenizer.convert_tokens_to_ids(tokens)
            segment_id = [1] * len(tokens)

            self.tokens_tensor = torch.tensor([idx])
            self.segments_tensors = torch.tensor([segment_id])

            with torch.no_grad():
                outputs = self.model(self.tokens_tensor, self.segments_tensors)
                hidden_states = outputs[2]

            return hidden_states[-2][0][main_token_id]

        except ValueError:
            raise ValueError(
                f'The word: "{main_word}" does not exist in the list of tokens: {tokens} from {doc}'
            )


class SenseRepresentation:
    """
    This class is used to recreate the sense representation for a given word in a given period of time.

    This implementation is based on the fine-grained sense representations with deep contextualized word embeddings,
    i.e., represent each sense as a distinguished sense embedding. We directly adopt the fine-grained senses
    defined by lexicographers. Comparing with existing diachronic sense studies, our method does not rely on
    human interpretations or mappings to dictionary definitions.

    For a sense \( s_j \) of word \( w_i \),
    we can obtain its example sentences \( \{ \text{Sent}_{wi}^{s_j} \} \) from a dictionary. After feeding them
    into a pre-trained language model, \( w_i \)'s token representations \( \{ e_{wi}^{s_j} \} \) can be retrieved
    from the final hidden layer of the model. The sense embedding \( e_{ij} \) of \( s_j \) is computed by taking
    the average of \( \{ e_{wi}^{s_j} \} \).

    Citation:
    \cite{author2024fine}
    # todo what citation is to be used here?
    """

    target_word: str

    def __init__(
            self,
            target_word: str,
            model_path: Union[str, Path] = None,

    ):
        """
        Initialize the class with the target word and the model path.
        Args:
            target_word: The main word to extract the sense representation for.
            model_path: The path to the pretrained model.
        """
        self.model = VectorEmbeddings.load(model_path)
        self.target_word = target_word

    def _infer_sentence_embedding(self, sense_examples: List[str]) -> torch.Tensor:
        """
        Infer the embeddings of the give_word in each example of the sense.
        Args:
            sense_examples: List of examples for the sense.
        Returns: torch.Tensor
        """
        for example in sense_examples:
            yield self.model.infer_vector(
                doc=example,
                main_word=self.target_word
            )

    def _get_senses(self) -> List[str]:
        """
        Get the senses for the target word.
        Returns: List[str]
        """
        synsets = wn.synsets(self.target_word)
        senses = [synset.definition() for synset in synsets]
        return senses

    def get_sentence_examples(self, sense: int, min_examples=5):
        """
        Extract or generate at least min_examples sentence examples for a specific sense of a word in WordNet.

        Args:
        - sense (int): The sense number of the word.
        - min_examples (int): Minimum number of examples required.

        Returns:
        - List[str]: A list of sentence examples.
        """
        synsets = wn.synsets(self.target_word)

        # Check if the sense number is valid
        if sense <= 0 or sense > len(synsets):
            raise ValueError(f"Sense number {sense} is out of range for the word '{self.target_word}'.")

        # Get the specific synset based on the sense number
        synset = synsets[sense - 1]

        # Retrieve the examples for the synset
        examples = synset.examples()

        # If the number of examples is less than the minimum required, generate more
        if len(examples) < min_examples:
            context_sentences = []
            for sentence in brown.sents():
                if self.target_word in sentence:
                    context_sentences.append(' '.join(sentence))
                if len(context_sentences) >= min_examples:
                    break

            # Add more examples from the Brown corpus if needed
            for context_sentence in context_sentences:
                examples.append(context_sentence)
                if len(examples) >= min_examples:
                    break

        return examples[:min_examples]

    def infer_representation(self) -> List[WordSenseEmbedding]:
        """
        Infer the mean vector embedding for the given word across all the examples in the senses.
        Returns: WordSenseEmbedding object.
        """
        senses = self._get_senses()
        sense_embeddings = []
        for idx, sense in tqdm(enumerate(senses),
                               desc=f'Processing the sense representation for the word: {self.target_word}',
                               total=len(senses)):
            examples = self.get_sentence_examples(idx + 1)
            try:
                msg.good(f'{"-" * 10} Processing the sense: {sense} {"-" * 10}')
                embeddings = list(self._infer_sentence_embedding(examples))
                sense_embedding = torch.mean(torch.stack(embeddings), dim=0).tolist()
                sense_embeddings += [
                    WordSenseEmbedding(
                        id=str(idx),
                        definition=sense,
                        embedding=sense_embedding
                    )
                ]
            except ValueError:
                continue
        return sense_embeddings



import pandas as pd
from .methods.lda_prototype import LDAPrototype
from .methods.rolling_lda import RollingLDA
from .preprocessing import chunk_creation
import warnings
from typing import Union, List, Tuple, Callable
from datetime import datetime
from .preprocessing.preprocess import preprocess


class TTTAPipeline():
    def __init__(self, corpus: Union[pd.DataFrame, str] = None, model: str = None, how: Union[str, List[datetime]] = "M", preprocessing: callable = None,
                 model_save_path: str = None, text_column: str = "text", date_column: str = "date", **kwargs) -> None:
        if corpus is not None:
            if isinstance(corpus, str):
                self.corpus = self._read_corpus_file(corpus)
            elif isinstance(corpus, pd.DataFrame):
                self.corpus = corpus
            else:
                raise TypeError("corpus must be a pandas DataFrame or a path to a corpus file!")
        else:
            self.corpus = None
            warnings.warn("No corpus file was given. A pipeline object will be created, but not be trained on any data.")
        if preprocessing is not None:
            self.corpus[text_column] = preprocessing(self.corpus)
        else:
            if model != "sense_disambiguation":
                self.corpus[text_column] = preprocess(self.corpus, **kwargs)

        if model is None:
            self.model = None
            warnings.warn("No model was given. A pipeline object will be created, but not be trained on any data.")
        elif model in ["LDA", "LDAPrototype"]:
            self.model = LDAPrototype(**kwargs)
        elif model == "RollingLDA":
            self.model = RollingLDA(how=how, **kwargs)
        elif model == "PoiRR":
            self.model = None
            warnings.warn("PoiRR is not yet implemented!")
        elif model == "diachronic_alignment":
            self.model = None
            warnings.warn("diachronic_alignment is not yet implemented!")
        elif model == "sense_disambiguation":
            self.model = None
            warnings.warn("sense_disambiguation is not yet implemented!")
        else:
            raise ValueError("model not recognized!")
        self.model.fit(self.corpus, text_column=text_column, date_column=date_column, **kwargs)
        if model_save_path is not None:
            self.model.save(model_save_path)

    def _read_corpus_file(self, path):
        """
        Reads a corpus file and returns a pandas DataFrame with the columns "date" and "text".
        Args:
            path: path to the corpus file
        Returns:
            pandas DataFrame with the columns "date" and "text"
        """
        if path.endswith('.csv', 'tsv'):
            df = pd.read_csv(path)
        elif path.endswith('.json'):
            df = pd.read_json(path)
        elif path.endswith('.pickle', 'pkl'):
            df = pd.read_pickle(path)
        elif path.endswith('.xml'):
            df = pd.read_xml(path)
        elif path.endswith('.hdf'):
            df = pd.read_hdf(path)
        elif path.endswith('.sql'):
            df = pd.read_sql(path)
        elif path.endswith('.xlsx', "xls"):
            df = pd.read_excel(path)
        else:
            raise ValueError(f'Unsupported filetype: {path.split(".")[-1]}')
        return df


import os
import shutil
import unittest
import numpy as np
import pandas as pd
import sys
sys.path.append('../src')
from ttta.preprocessing.chunk_creation import _get_time_indices

class TestChunkCreation(unittest.TestCase):
    def setUp(self):

        self.texts = [["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"],
                      ["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"]]
        self.dates = [pd.to_datetime("2021-01-01"), pd.to_datetime("2021-01-02"), pd.to_datetime("2021-01-03")] * 7
        self.df = pd.DataFrame({"text": self.texts, "date": self.dates})
        self.df.sort_values("date", inplace=True)

    def test_time_indices(self):
        _get_time_indices(self.df, how="1D", last_date=pd.to_datetime("2020-12-31"), date_column="date", min_docs_per_chunk=1.0)
        _get_time_indices(self.df, how=[pd.to_datetime("2021-01-01"), pd.to_datetime("2021-01-02"), pd.to_datetime("2021-01-03")],
                          last_date=pd.to_datetime("2020-12-31"), date_column="date", min_docs_per_chunk=1.0)

    def test_time_indeces_invalid(self):
        with self.assertRaises(TypeError):
            _get_time_indices(1)
        with self.assertRaises(TypeError):
            _get_time_indices(self.df, 1)
        with self.assertRaises(TypeError):
            _get_time_indices(self.df, last_date=1)
        with self.assertRaises(ValueError):
            _get_time_indices(self.df, last_date=pd.to_datetime("2021-12-31"))
        with self.assertRaises(TypeError):
            _get_time_indices(self.df, date_column=1)
        with self.assertRaises(ValueError):
            _get_time_indices(self.df, min_docs_per_chunk=-1)
        with self.assertRaises(TypeError):
            _get_time_indices(self.df, min_docs_per_chunk=1.5)




import os
import unittest
import numpy as np
import pandas as pd
import sys
sys.path.append('../src')
from ttta.methods.lda_prototype import LDAPrototype

class TestLDAPrototype(unittest.TestCase):

    def setUp(self):
        self.texts = [["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"]]
        self.total_words = len([word for doc in self.texts for word in doc])
        self.unique_words = len(set([word for doc in self.texts for word in doc]))
    def test_init_with_valid_parameters(self):
        lda = LDAPrototype(K=5.0, alpha=0.2, gamma=0.2, prototype=3.0, topic_threshold=[5.0, 0.002], prototype_measure="jaccard",
                           min_count=2.0, max_assign=False, verbose=1)
        self.assertIsInstance(lda, LDAPrototype)
        self.assertEqual(lda._K, 5)
        self.assertEqual(lda._alpha.shape[0], 5)
        self.assertEqual(lda._gamma.shape[0], 5)
        self.assertEqual(lda._prototype, 3)
        self.assertEqual(lda._threshold, [5, 0.002])
        self.assertEqual(lda._measure, "jaccard")
        self.assertEqual(lda._min_count, 2)
        self.assertFalse(lda._max_assign)
        self.assertEqual(lda._verbose, 1)

    def test_init_invalid_parameters(self):
        with self.assertRaises(TypeError):
            lda = LDAPrototype()
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=0)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K="K")
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=1.5)
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, alpha=0)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, alpha="alpha")
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, alpha=-0.5)
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, alpha=np.array([1, 1, 1]))
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, alpha=np.array([-1, 1, 1, 1, 1]))
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, gamma=-0.5)
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, gamma=np.array([1, 1, 1]))
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, gamma=np.array([-1, 1, 1, 1, 1]))
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, gamma=0)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, gamma="gamma")
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, prototype=0)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, prototype="prototype")
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, prototype=1.5)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, topic_threshold=0)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, topic_threshold=[0.1, 0.1])
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, topic_threshold=[1, "a"])
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, topic_threshold=[1, 1.0])
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, prototype_measure=0)
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, prototype_measure="prototype_measure")
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, min_count="min_count")
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, min_count=3.5)
        with self.assertRaises(ValueError):
            lda = LDAPrototype(K=5, min_count=-1)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, max_assign=3)
        with self.assertRaises(TypeError):
            lda = LDAPrototype(K=5, verbose="verbose")

    def test_fit_with_valid_input(self):
        lda = LDAPrototype(K=3, prototype=1, min_count=0, verbose=2, max_assign=True)
        lda.fit([[0, 1, 2, 3], [4, 5, 6, 7, 8]], epochs=10.0, workers=1.0, first_chunk=True, chunk_end=float(len(self.texts)), memory_start=1.0)
        self.assertTrue(lda._is_trained)
        self.assertIsNotNone(lda._assignments)
        self.assertIsNotNone(lda._word_vec)
        self.assertIsNotNone(lda._document_topic_matrix)

    def test_fit_invalid_input(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        with self.assertRaises(TypeError):
            lda.fit(0, epochs=10)
        with self.assertRaises(TypeError):
            lda.fit(np.array([0]), epochs=10)
        with self.assertRaises(TypeError):
            lda.fit("test", epochs=10)
        with self.assertRaises(TypeError):
            lda.fit([0, 0], epochs=10)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, epochs="epochs")
        with self.assertRaises(ValueError):
            lda.fit(self.texts, epochs=-1)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, epochs=1.5)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, first_chunk=-1)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, chunk_end="chunk_end")
        with self.assertRaises(ValueError):
            lda.fit(self.texts, chunk_end=-1)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, chunk_end=1.5)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, memory_start="memory_start")
        with self.assertRaises(ValueError):
            lda.fit(self.texts, memory_start=-1)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, memory_start=1.5)
        with self.assertRaises(TypeError):
            lda.fit(self.texts, workers="workers")
        with self.assertRaises(TypeError):
            lda.fit(self.texts, workers=1.5)
    def test_top_words(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda2 = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        lda2.fit(self.texts, epochs=10)
        top_words = lda.top_words(topic=1.0, number=2.0, importance=False)
        self.assertIsInstance(top_words, list)
        self.assertEqual(len(top_words), 2)
        top_words = lda.top_words(number=2.0, return_as_data_frame=False)
        self.assertIsInstance(top_words, list)
        self.assertEqual(len(top_words), 3)
        self.assertIsInstance(top_words[0], list)
        self.assertEqual(len(top_words[0]), 2)
        top_words = lda.top_words(number=2)
        top_words2 = lda2.top_words(number=2)
        self.assertIsInstance(top_words, pd.DataFrame)
        self.assertEqual(len(top_words), 2)
        self.assertEqual(len(top_words.columns), 3)
        self.assertFalse(top_words.equals(top_words2))
        top_words = lda.top_words(topic=1.0, number=2.0, importance=False)

    def test_top_words_invalid_input(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0, topic_threshold=[1, 0.002])
        lda2 = LDAPrototype(K=3, prototype=3, min_count=0, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(TypeError):
            top_words = lda.top_words(topic=3.5)
        with self.assertRaises(ValueError):
            top_words = lda.top_words(number=0)
        with self.assertRaises(TypeError):
            top_words = lda.top_words(number="number")
        with self.assertRaises(TypeError):
            top_words = lda.top_words(number=3.5)
        with self.assertRaises(TypeError):
            top_words = lda.top_words(topic="topic")
        with self.assertRaises(ValueError):
            top_words = lda.top_words(topic=0)
        with self.assertRaises(ValueError):
            top_words = lda.top_words(topic=4)
        with self.assertRaises(TypeError):
            top_words = lda.top_words(importance="importance")
        with self.assertRaises(TypeError):
            top_words = lda.top_words(word_topic_matrix="word_topic_matrix")
        with self.assertRaises(TypeError):
            top_words = lda.top_words(return_as_data_frame=-1)
        with self.assertRaises(AttributeError):
            top_words = lda2.top_words()

    def test_get_word_and_doc_vector(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        word_vec, doc_vec = lda._get_word_and_doc_vector(lda._dtm)
        self.assertIsInstance(word_vec, np.ndarray)
        self.assertIsInstance(doc_vec, np.ndarray)
        self.assertEqual(word_vec.shape, doc_vec.shape)

    def test_get_word_and_doc_vector_invalid_input(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(TypeError):
            word_vec, doc_vec = lda._get_word_and_doc_vector(0)
        with self.assertRaises(TypeError):
            word_vec, doc_vec = lda._get_word_and_doc_vector()

    def test_get_word_topic_matrix(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        word_topic_matrix = lda.get_word_topic_matrix()
        self.assertIsInstance(word_topic_matrix, np.ndarray)
        self.assertEqual(word_topic_matrix.shape, (self.unique_words, 3))

    def test_get_word_topic_matrix_invalid_input(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(ValueError):
            word_topic_matrix = lda.get_word_topic_matrix(np.ndarray([0]), lda.get_assignment_vec())
        with self.assertRaises(ValueError):
            word_topic_matrix = lda.get_word_topic_matrix(0, lda.get_assignment_vec())
        with self.assertRaises(TypeError):
            word_topic_matrix = lda.get_word_topic_matrix("a", lda.get_assignment_vec())
        with self.assertRaises(TypeError):
            word_topic_matrix = lda.get_word_topic_matrix(lda.get_word_vec(), "a")

    def test_get_assignment_vec(self):
        lda = LDAPrototype(K=3, prototype=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        assignment_vec = lda.get_assignment_vec()
        self.assertIsInstance(assignment_vec, np.ndarray)
        self.assertEqual(assignment_vec.shape, (self.total_words,))

    def test_get_word_vec(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        word_vec = lda.get_word_vec()
        self.assertIsInstance(word_vec, np.ndarray)
        self.assertEqual(word_vec.shape, (self.total_words,))

    def test_get_doc_vec(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        doc_vec = lda.get_doc_vec()
        self.assertIsInstance(doc_vec, np.ndarray)
        self.assertEqual(doc_vec.shape, (self.total_words,))
    def test_get_document_topic_matrix(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        document_topic_matrix = lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec())
        self.assertIsInstance(document_topic_matrix, np.ndarray)
        self.assertEqual(document_topic_matrix.shape, (len(self.texts), 3))

    def test_get_document_topic_matrix_invalid_input(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(ValueError):
            document_topic_matrix = lda.get_document_topic_matrix(0, lda.get_assignment_vec())
        with self.assertRaises(TypeError):
            document_topic_matrix = lda.get_document_topic_matrix(lda.get_doc_vec(), "a")
        with self.assertRaises(TypeError):
            document_topic_matrix = lda.get_document_topic_matrix("a", lda.get_assignment_vec())

    def test_get_vocab(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        vocab = lda.get_vocab()
        self.assertIsInstance(vocab, list)
        self.assertEqual(len(vocab), self.unique_words)

    def test_get_params(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        params = lda.get_params()
        self.assertIsInstance(params, dict)
        self.assertEqual(params["_K"], 3)
        self.assertTrue(np.array_equal(params["_alpha"], [1/3] * 3))
        self.assertTrue(np.array_equal(params["_gamma"], [1/3] * 3))
        self.assertEqual(params["_prototype"], 10)
        self.assertEqual(params["_threshold"], [5, 0.002])
        self.assertEqual(params["_measure"], "jaccard")
        self.assertEqual(params["_min_count"], 0)
        self.assertFalse(params["_max_assign"], False)
        self.assertEqual(params["_verbose"], 1)

    def test_set_params(self):
        lda = LDAPrototype(K=3, min_count=0)
        lda.fit(self.texts, epochs=10)
        lda.set_params({"_K": 5, "_alpha": np.array([0.5] * 5), "_gamma": np.array([0.5] * 5), "_prototype": 2, "_threshold": [3, 0.002], "_measure": "cosine",
                        "_min_count": 3, "_max_assign": True, "_verbose": 0})
        self.assertEqual(lda._K, 5)
        self.assertTrue(np.array_equal(lda._alpha, [0.5] * 5))
        self.assertTrue(np.array_equal(lda._gamma, [0.5] * 5))
        self.assertEqual(lda._prototype, 2)
        self.assertEqual(lda._threshold, [3, 0.002])
        self.assertEqual(lda._measure, "cosine")
        self.assertEqual(lda._min_count, 3)
        self.assertTrue(lda._max_assign)
        self.assertEqual(lda._verbose, 0)
        with self.assertRaises(TypeError):
            lda.set_params(0)

    def test_save_and_load(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        lda.save("lda_test.pickle")
        lda2 = LDAPrototype(K=3, min_count=0)
        lda2.load("lda_test.pickle")
        os.remove("lda_test.pickle")

    def test_save_and_load_invalid_input(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(TypeError):
            lda.save(0)
        with self.assertRaises(TypeError):
            lda.load(0)

    def test_shrink_model(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        with self.assertRaises(NotImplementedError):
            lda.shrink_model(2)

    def test_is_trained(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        self.assertFalse(lda.is_trained())
        lda.fit(self.texts, epochs=10)
        self.assertTrue(lda.is_trained())

    def test_create_lda_parameters(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        self.assertCountEqual(lda._create_lda_parameters(0.5), [0.5, 0.5, 0.5])
        self.assertCountEqual(lda._create_lda_parameters(np.array([1, 1, 1])), [1, 1, 1])

    def test_create_lda_parameters_invalid_input(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        with self.assertRaises(TypeError):
            lda._create_lda_parameters("a")
        with self.assertRaises(ValueError):
            lda._create_lda_parameters(-1)
        with self.assertRaises(ValueError):
            lda._create_lda_parameters([1, 2, 3, 4])
        with self.assertRaises(ValueError):
            lda._create_lda_parameters(np.array([-1, 2, 4]))

    def test_wordclouds(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        lda.wordclouds(None, 1.0, height=100.0, width=100.0, show=False)
        lda.wordclouds(1.0, 5.0, None, height=100.0, width=100.0, show=True, word_topic_matrix=lda.get_word_topic_matrix(lda.get_word_vec(), lda.get_assignment_vec()))
        os.remove("wordclouds.pdf")

    def test_wordclouds_invalid_input(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(TypeError):
            lda.wordclouds(1.5)
        with self.assertRaises(ValueError):
            lda.wordclouds(-1)
        with self.assertRaises(ValueError):
            lda.wordclouds(number=-1)
        with self.assertRaises(TypeError):
            lda.wordclouds(number=0.5)
        with self.assertRaises(TypeError):
            lda.wordclouds(path=0.5)
        with self.assertRaises(ValueError):
            lda.wordclouds(width=-1)
        with self.assertRaises(TypeError):
            lda.wordclouds(width=0.5)
        with self.assertRaises(ValueError):
            lda.wordclouds(height=-1)
        with self.assertRaises(TypeError):
            lda.wordclouds(height=0.5)
        with self.assertRaises(TypeError):
            lda.wordclouds(show=0.5)
        with self.assertRaises(ValueError):
            lda.wordclouds(path=None, show=False)

    def test_calculate_prototype_invalid_input(self):
        lda = LDAPrototype(K=3, min_count=0, prototype=3, topic_threshold=[1, 0.002])
        lda.fit(self.texts, epochs=10)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(0.5, lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), 0, lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), 0, lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), 0, np.array(1), 1, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), 0.5, 1, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1.5, 10, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10.5, 0)
        with self.assertRaises(TypeError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10, 0.5)
        with self.assertRaises(ValueError):
            lda._calculate_prototype(lda.get_word_vec(), lda.get_doc_vec(), lda.get_word_topic_matrix(), lda.get_document_topic_matrix(lda.get_doc_vec(), lda.get_assignment_vec()), np.array(1), 1, 10, 0-1)
if __name__ == '__main__':
    unittest.main()



import os
import shutil
import unittest
import numpy as np
import pandas as pd
import sys
sys.path.append('../src')
from ttta.methods.topic_matching import TopicClusters
from ttta.methods.lda_prototype import LDAPrototype

class TestTopicMatching(unittest.TestCase):
    def setUp(self):
        self.texts = [["lorem", "ipsum", "dolor", "sit", "amet", "consectetur", "adipiscing", "elit"],
                      ["sed", "do", "eiusmod", "tempor", "incididunt", "ut", "labore", "et", "dolore", "magna"],
                      ["aliqua", "ut", "enim", "ad", "minim", "veniam", "quis", "nostrud", "exercitation", "ullamco"],
                      ["laboris", "nisi", "ut", "aliquip", "ex", "ea", "commodo", "consequat"],
                      ["duis", "aute", "irure", "dolor", "in", "reprehenderit", "in", "voluptate", "velit", "esse"],
                      ["cillum", "dolore", "eu", "fugiat", "nulla", "pariatur"],
                      ["excepteur", "sint", "occaecat", "cupidatat", "non", "proident", "sunt", "in", "culpa", "qui"],
                      ["officia", "deserunt", "mollit", "anim", "id", "est", "laborum"]]
        self.lda = LDAPrototype(K=3, prototype=1, min_count=0, verbose=2, max_assign=True)
        self.lda.fit(self.texts, epochs=10.0, workers=1.0, first_chunk=True, chunk_end=float(len(self.texts)), memory_start=1.0)
        self.matrices = [self.lda.get_word_topic_matrix(), self.lda.get_word_topic_matrix()]

    def test_init(self):
        model = TopicClusters(self.matrices, K=3.0, topic_threshold=[1, 0.01])
        model.get_matched_topics()
        with self.assertRaises(TypeError):
            TopicClusters(1)
        with self.assertRaises(ValueError):
            TopicClusters([self.lda.get_word_topic_matrix()])
        with self.assertRaises(TypeError):
            TopicClusters([1])
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, measure="a")
        with self.assertRaises(TypeError):
            TopicClusters(self.matrices, measure=1)
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, K=1)
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, K=4)
        with self.assertRaises(TypeError):
            TopicClusters(self.matrices, K=3.5)
        with self.assertRaises(TypeError):
            TopicClusters(self.matrices, topic_threshold=1)
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, topic_threshold=[-1, 0.1])
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, topic_threshold=[1, -0.1])
        with self.assertRaises(ValueError):
            TopicClusters(self.matrices, topic_threshold=[1, 1.1])

    def test_cluster_topics(self):
        model = TopicClusters(self.matrices, K=3.0, topic_threshold=[1, 0.01])
        with self.assertRaises(TypeError):
            model._cluster_topics("a", 1, "cosine", dict())
        with self.assertRaises(TypeError):
            model._cluster_topics(1.5, 1, "cosine", dict())
        with self.assertRaises(TypeError):
            model._cluster_topics(1, "a", "cosine", dict())
        with self.assertRaises(TypeError):
            model._cluster_topics(1, 1.5, "cosine", dict())
        with self.assertRaises(TypeError):
            model._cluster_topics(1, 1, 1, dict())
        with self.assertRaises(TypeError):
            model._cluster_topics(1, 1, "cosine", [])


import unittest
from unittest.mock import patch, MagicMock
import torch
from pathlib import Path
from src.ttta.preprocessing.schemas import WordSenseEmbedding
from src.ttta.methods.sense_representation import VectorEmbeddings, SenseRepresentation


# Assuming the classes VectorEmbeddings and SenseRepresentation are defined as shown above

class TestVectorEmbeddings(unittest.TestCase):

    @patch('transformers.BertModel.from_pretrained')
    @patch('transformers.BertTokenizer.from_pretrained')
    def setUp(self, mock_tokenizer, mock_model):
        self.mock_tokenizer = mock_tokenizer
        self.mock_model = mock_model
        self.mock_tokenizer.return_value = MagicMock()
        self.mock_model.return_value = MagicMock()
        self.model_path = "bert-base-uncased"
        self.vector_embeddings = VectorEmbeddings(self.model_path)

    def test_initialization(self):
        self.assertIsInstance(self.vector_embeddings, VectorEmbeddings)
        self.assertEqual(self.vector_embeddings.model_path, Path(self.model_path))
        self.assertTrue(self.vector_embeddings.vocab)

    def test_infer_vector(self):
        doc = "The bank will not provide the loan."
        main_word = "bank"

        self.vector_embeddings.bert_tokenizer.tokenize.return_value = ["[CLS]", "the", "bank", "will", "not", "provide",
                                                                       "the", "loan", ".", "[SEP]"]
        self.vector_embeddings.bert_tokenizer.convert_tokens_to_ids.return_value = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        output_tensor = torch.rand((1, 10, 768))
        self.vector_embeddings.model.return_value = (None, None, [output_tensor])

        inferred_vector = self.vector_embeddings.infer_vector(doc, main_word)
        self.assertIsInstance(inferred_vector, torch.Tensor)

    def test_load(self):
        vector_embeddings = VectorEmbeddings.load(self.model_path)
        self.assertIsInstance(vector_embeddings, VectorEmbeddings)
        self.assertEqual(vector_embeddings.model_path, Path(self.model_path))


class TestSenseRepresentation(unittest.TestCase):

    @patch('transformers.BertModel.from_pretrained')
    @patch('transformers.BertTokenizer.from_pretrained')
    def setUp(self, mock_tokenizer, mock_model):
        self.mock_tokenizer = mock_tokenizer
        self.mock_model = mock_model
        self.mock_tokenizer.return_value = MagicMock()
        self.mock_model.return_value = MagicMock()
        self.target_word = "bank"
        self.model_path = "bert-base-uncased"
        self.sense_representation = SenseRepresentation(self.target_word, self.model_path)

    def test_initialization(self):
        self.assertIsInstance(self.sense_representation, SenseRepresentation)
        self.assertEqual(self.sense_representation.target_word, self.target_word)
        self.assertTrue(self.sense_representation.model.vocab)

    @patch('nltk.corpus.wordnet.synsets')
    def test_get_senses(self, mock_synsets):
        mock_synset = MagicMock()
        mock_synset.definition.return_value = "A financial institution that accepts deposits and channels the money into lending activities."
        mock_synsets.return_value = [mock_synset]

        senses = self.sense_representation._get_senses()
        self.assertEqual(len(senses), 1)
        self.assertEqual(senses[0], mock_synset.definition())

    @patch('nltk.corpus.wordnet.synsets')
    @patch('nltk.corpus.brown.sents')
    def test_get_sentence_examples(self, mock_brown_sents, mock_synsets):
        mock_synset = MagicMock()
        mock_synset.examples.return_value = ["The bank will not provide the loan."]
        mock_synsets.return_value = [mock_synset]
        mock_brown_sents.return_value = [["The", "bank", "is", "near", "the", "river"]]

        examples = self.sense_representation.get_sentence_examples(1, min_examples=5)
        self.assertGreaterEqual(len(examples), 5)

    @patch('torch.mean')
    @patch('SenseRepresentation._infer_sentence_embedding')
    @patch('SenseRepresentation._get_senses')
    def test_infer_representation(self, mock_get_senses, mock_infer_sentence_embedding, mock_mean):
        mock_get_senses.return_value = [
            "A financial institution that accepts deposits and channels the money into lending activities."]
        mock_infer_sentence_embedding.return_value = [torch.rand((768,))]
        mock_mean.return_value = torch.rand((768,))

        embeddings = self.sense_representation.infer_representation()
        self.assertIsInstance(embeddings, list)
        self.assertGreater(len(embeddings), 0)
        self.assertIsInstance(embeddings[0], WordSenseEmbedding)


if __name__ == '__main__':
    unittest.main()




import os
import json
from numpy.linalg import norm
from ttta.preprocessing.schemas import Embedding, WordSenseEmbedding, SenseEmbedding
from ttta.preprocessing.settings import EmbeddingFiles, FileLoader
import logging
import numpy as np
from collections import Counter
from typing import List
from pydantic import parse_obj_as

class Similarities:
    """
    Class that computes the similarity between the embeddings of the words and the embeddings of the senses

    Attributes
    ----------
        files: EmbeddingFiles
            EmbeddingFiles object that contains the paths to the files
        embedding_component: Embedding
            Embedding object that contains the word and the embeddings
        sense_component: SenseEmbedding
            SenseEmbedding object that contains the sense id and the embeddings
        embeddings_senses: List[Dict]
            List of dictionaries that contains the word and the senses
        words: str
            The list of all the words in the embeddings_senses file
        embeddings_examples: List[Dict]
            List of dictionaries that contains the word and the embeddings

    Methods
    -------
        _lookup_examples_that_match(words: str) -> List[Embedding]
            Returns the embeddings of the words that are present in the embeddings_examples file
        _search_word_sense(main_word: str) -> List[SenseEmbedding]
            Returns the senses of the word that are present in the embeddings_senses file
        _cos_sim(vect_a: np.array, vect_b: np.array) -> float
            Returns the cosine similarity between two vectors
        __call__(main_word: str, year: int, path_embeddings_file: str) -> None
            Computes the similarity between the embeddings of the words and the embeddings of the senses

    """
    def __init__(
            self,
        ):

        self.files = EmbeddingFiles()
        self.embedding_component = Embedding()
        self.embeddings_senses, self.words = FileLoader.load_files(self.__class__.__name__)

        self.embeddings_examples = None
        self.root_dir = self.files.embeddings_root_dit


    def _lookup_examples_that_match(self) -> List[Embedding]:
        """
        Returns the embeddings of the words that are present in the embeddings_examples file
        Returns:
            List[Embedding]

        """
        if self.embeddings_examples is None:
            raise ValueError(
                'Embedding examples not initialized'
            )

        embeddings = [embed.word for embed in self.embeddings_examples]
        for word in list(set(embeddings) & set(self.words.split('\n'))):
            yield next(e for e in self.embeddings_examples if word == e.word)


    def _search_word_sense(self, main_word:str) -> List[WordSenseEmbedding]:
        """
        Returns the senses of the word that are present in the embeddings_senses file
        Args:
            main_word: str

        Returns:
            List[SenseEmbedding]
        """
        for w in self.embeddings_senses:
            if not w.word == main_word:
                continue

            return w.senses

    def _cos_sim(self, vect_a:np.array, vect_b:np.array):
        """
        Returns the cosine similarity between two vectors
        Args:
            vect_a: np.array
            vect_b: np.array

        Returns:
            float
        """
        return (vect_a @ vect_b)/(norm(vect_a) * norm(vect_b))

    def __call__(self, main_word:str, year:int, path_embeddings_file:str):
        w_senses = self._search_word_sense(main_word)
        if len(w_senses) == 0:
            raise TypeError(f'The word {main_word} is not present in the embeddings_senses file')

        print(f'{"-" * 10} Computing Similarities for the word {main_word} {"-" * 10}')

        with open(os.path.join(self.root_dir, path_embeddings_file), 'r') as f:
            logging.info(f'{"-" * 10} Loading the embeddings examples file: {f.name} {"-" * 10}')
            self.embeddings_examples = parse_obj_as(List[Embedding], json.load(f))

        self.all_embeddings = list(self._lookup_examples_that_match())

        all_sims = []
        for embed in self.all_embeddings:
            for ex in embed.embeddings:
                s_argmax = np.argmax([self._cos_sim(np.array(ex), np.array(s.embedding)) for s in w_senses])
                all_sims += [w_senses[s_argmax].id]

        return list(map(lambda x: x[1]/len(all_sims), Counter(all_sims).most_common()))


def sim_on_all_words():
    """
    Computes the similarity between the embeddings of the words and the embeddings of the senses
    Returns:
        None

    """
    sim = Similarities()
    word_proportions = {}

    with open(sim.files.poly_words_f, 'r') as f:
        words = f.read()

    for w_ in words.split('\n'):
        s = []
        w_senses = sim._search_word_sense(w_)
        word_proportions['word'] = w_
        word_proportions['sense_ids'] = [w.id for w in w_senses]
        for year in sim.files.years_used:
            s += [{'year': year, 'props': sim(w_, year, path_embeddings_file=f'embeddings_{year}.json')}.copy()]

        word_proportions['props'] = s
        with open(f'../embeddings_similarity/embeddings_sim_{w_}.json', 'w') as f:
            json.dump(word_proportions, f, indent=4)


if __name__ == '__main__':
    sim_on_all_words()




import os

from pydantic_settings import BaseSettings
from pydantic import Field
from dotenv import load_dotenv
import logging
import json
from pydantic import parse_obj_as
from typing import List
# from src.ttta.diachronic_embeddings.utils.components import Words, SenseEmbedding

load_dotenv(verbose=True)


class OxfordAPISettings(BaseSettings):
    accept: str = Field(..., env="ACCEPT")
    app_id: str = Field(..., env="APP_ID")
    app_key: str = Field(..., env="APP_KEY")
    url: str = Field(..., env="URL")
    strict_match: str = Field(..., env="STRICT_MATCH")

    def __str__(self):
        return self.accept + '\n' + self.app_id + '\n' + self.app_key + '\n' + self.url + '\n'

    class Config:
        env_file = "../.env"
        env_file_encoding = "utf-8"


class EmbeddingFiles(BaseSettings):
    poly_words_f:str = os.getenv('POLY_WORDS')
    oxford_word_senses: str = os.getenv('OXFORD_WORD_SENSES')
    sense_embeddings: str = os.getenv('SENSE_EMBEDDINGS')
    embeddings_root_dit:str = '../embeddings'
    bert_model:str = 'bert_model_new'
    years_used:list = [1980, 1982, 1985, 1987, 1989, 1990, 1995, 2000, 2001, 2002, 2003, 2005, 2008, 2009, 2010, 2012,
                     2013, 2015, 2016, 2017, 2018]

# class FileLoader():
#     @classmethod
#     def load_files(
#             cls,
#             module:str,
#             file_keyword:str=None
#         ):
#         if not module in ['Similarities', 'ExtractSenseEmbeddings', 'PolynomialFitting']:
#             raise ValueError(
#                 f'The module {module} does not exist'
#             )
#
#         logging.basicConfig(level=logging.NOTSET)
#         files = EmbeddingFiles()
#
#         with open(files.poly_words_f, 'r') as f:
#             words = f.read()
#
#         if module == 'Similarities':
#             with open(files.sense_embeddings, 'r') as f:
#                 logging.info(f'{"-" * 10} Loading the embeddings for senses: {f.name} {"-" * 10}')
#                 senses_embeds = parse_obj_as(List[SenseEmbedding], json.load(f))
#
#             return senses_embeds, words
#
#         if module == 'ExtractSenseEmbeddings':
#             with open(files.oxford_word_senses) as f: all_words = parse_obj_as(List[Words], json.load(f))
#             return all_words
#
#         if module == 'PolynomialFitting':
#             if not file_keyword in words.split('\n'):
#                 raise ValueError(
#                     f'The specified keyword: "{file_keyword}" does not apply to any similarity embedding file'
#                 )
#
#             with open(f'../embeddings_similarity/embeddings_sim_{file_keyword}.json') as f:
#                 word_props = json.load(f)
#
#             return word_props, words










import pickle
import warnings
from multiprocessing.pool import ThreadPool as Pool
import faulthandler
import matplotlib
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from wordcloud import WordCloud
from scipy.sparse import csr_array
from .LDA.lda_gibbs import vanilla_gibbs_func, load_wk_mat_func, final_assignment_func, load_dk_mat_func
from .LDA.importance_calculation import calculate_importance
from typing import Union, List, Tuple, Callable
from tqdm import tqdm
from matplotlib.backends.backend_pdf import PdfPages
from .topic_matching import TopicClusters
from ..preprocessing.preprocess import create_dtm, get_word_and_doc_vector
faulthandler.enable()


class LDAPrototype:
    def __init__(self, K: int, alpha: float = None, gamma: float = None, prototype: int = 10, topic_threshold: List[Union[int, float]] = None,
                 prototype_measure: Union[str, Callable] = "jaccard", min_count: int = 2, max_assign: bool = False, verbose: int = 1):
        """
        Implements the LDAPrototype model, which trains multiple LDA models and selects the best one based on similarity measures.
        The LDAs are trained in C.
        Args:
            K: Number of topics
            alpha: alpha parameter for the LDA
            gamma: gamma parameter for the LDA
            prototype: Number of prototypes to train. If 1, the model is equivalent to a normal LDA.
            topic_threshold: Threshold for the minimum number of occurrences in a topic for a word to be relevant for the prototype selection
            prototype_measure: Measure to use for prototype selection
            min_count: Minimum number of occurrences for a word to be included in the vocabulary
            max_assign: Should the final assignment of the trained LDAs be chosen by maximum assignment from all previous training iterations
                        or should it be chosen randomly based on the word-topic probabilities (default)?
            verbose: Verbosity level. 0 does not print anything, 1 prints runtime-relevant information, 2 and higher
                     shows debugging information.
        """
        if not isinstance(K, int):
            try:
                if K == int(K):
                    K = int(K)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("K must be an integer!")
        if K < 2:
            raise ValueError("K must be a natural number greater than 1!")
        if not isinstance(alpha, float) and not isinstance(alpha, np.ndarray) and alpha is not None:
            try:
                alpha = float(alpha)
            except ValueError:
                try:
                    alpha = np.array(alpha, dtype=float)
                except ValueError:
                    raise TypeError("alpha must be a float or an array of length K!")
        if isinstance(alpha, np.ndarray):
            if alpha.shape[0] != K:
                raise ValueError("alpha must be a float or an array of length K!")
            if np.any(alpha < 0):
                raise ValueError("alpha must only contain floats greater than 0!")
        if isinstance(alpha, float) and alpha < 0:
            raise ValueError("alpha must be a float greater than 0!")
        if not isinstance(gamma, float) and not isinstance(gamma, np.ndarray) and gamma is not None:
            try:
                gamma = float(gamma)
            except ValueError:
                try:
                    gamma = np.array(gamma, dtype=float)
                except ValueError:
                    raise TypeError("gamma must be a float or an array of length K!")
        if isinstance(gamma, np.ndarray):
            if gamma.shape[0] != K:
                raise ValueError("gamma must be a float or an array of length K!")
            if np.any(gamma < 0):
                raise ValueError("gamma must only contain floats greater than 0!")
        if isinstance(gamma, float) and gamma < 0:
            raise ValueError("gamma must be a float greater than 0!")
        if not isinstance(prototype, int):
            try:
                if prototype != int(prototype):
                    raise ValueError
                prototype = int(prototype)
            except ValueError:
                raise TypeError("prototype must be an integer!")
        if prototype < 1:
            raise ValueError("prototype must be a natural number greater than 0")
        if topic_threshold is None:
            topic_threshold = [5, 0.002]
        elif not isinstance(topic_threshold, list):
            raise TypeError("topic_threshold must be a list of an integer and a float!")
        if not isinstance(topic_threshold[0], int):
            try:
                new_thresh = int(topic_threshold[0])
                if topic_threshold[0] != new_thresh:
                    raise ValueError
            except ValueError:
                raise TypeError("The first value in topic_threshold must be an integer!")
            topic_threshold[0] = new_thresh
        if not isinstance(topic_threshold[1], float):
            raise TypeError("The second value in topic_threshold must be a float!")
        if topic_threshold[0] < 0 or topic_threshold[1] < 0 or topic_threshold[1] >= 1:
            raise ValueError("topic_threshold must contain numbers larger or equal to 0. The second number must also be smaller than 1!")
        if not isinstance(prototype_measure, str) and not callable(prototype_measure):
            raise TypeError("prototype_measure must be a string or a callable!")
        elif prototype_measure not in ["jaccard", "cosine"] and not callable(prototype_measure):
            raise ValueError("prototype_measure must be either 'jaccard', 'cosine' or a callable!")
        if not isinstance(min_count, int):
            try:
                if min_count != int(min_count):
                    raise ValueError
                min_count = int(min_count)
            except ValueError:
                raise TypeError("min_count must be an integer!")
        if min_count < 0:
            raise ValueError("min_count must be a natural number!")
        if not isinstance(max_assign, bool):
            raise TypeError("max_assign must be a boolean!")
        if not isinstance(verbose, float):
            try:
                verbose = float(verbose)
            except ValueError:
                raise TypeError("verbose must be a float!")
        self._K = K
        self._alpha = self._create_lda_parameters(alpha)
        self._gamma = self._create_lda_parameters(gamma)
        self._prototype = prototype
        self._max_assign = max_assign
        self._verbose = verbose
        self._assignments = None
        self._vocab = []
        self._dtm = None
        self._word_vec = None
        self._doc_vec = None
        self._document_topic_matrix = []
        self._threshold = topic_threshold
        self._min_count = min_count
        self._measure = prototype_measure
        self._is_trained = False
        self._trained_words = [0]
        self._len_of_docs = []
        self._deleted_indices = []
        self._last_end = [0]

    def _create_lda_parameters(self, param: Union[np.ndarray, float, None]) -> np.ndarray:
        """
        Creates a parameter List for LDA from a float, a numpy array or automatically from K
        Args:
            param: parameter - either alpha or gamma
        Returns:
            param: parameter list
        """
        if not isinstance(param, float) and not isinstance(param, np.ndarray) and param is not None:
            try:
                param = float(param)
            except:
                try:
                    param = np.array(param, dtype=float)
                except ValueError:
                    raise TypeError("param must be a float or an array of length K!")
        if isinstance(param, float) and param <= 0:
            raise ValueError("param must be a positive float!")
        if isinstance(param, np.ndarray):
            if param.shape[0] != self._K:
                raise ValueError("param must be a float or an array of length K!")
            if np.any(param <= 0):
                raise ValueError("param must only contain floats greater than 0!")
        if param is None:
            param = np.repeat(1 / self._K, self._K)
        elif isinstance(param, float):
            param = np.repeat(param, self._K)
        else:
            param = param.astype(float)
        return param

    def _create_dtm(self, texts: List[List[str]]) -> None:
        """
        Creates a document-term matrix from a list of texts and updates the existing dtm if there is one.
        Stores both the dtm and the vocabulary in the class.
        Args:
            texts: list of texts
        Returns:
            None
        """
        if self._verbose > 0:
            print("Creating document-term matrix...")
        self._dtm, self._vocab, self._deleted_indices = create_dtm(texts, self._vocab, self._min_count, self._deleted_indices, self._dtm)
        if self._verbose > 1:
            print(f"Created document-term matrix with shape {self._dtm.shape}")

    @staticmethod
    def _get_word_and_doc_vector(dtm: Union[csr_array, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Turns a document-term matrix into index vectors. The word vector contains the vocabulary index for each word
        occurrence including multiple occurrences in one text. The document vector contains the document index for each
        word occurrence including multiple occurrences in one document.
        Args:
            dtm: document-term matrix
        Returns:
            word_vec: word vector
            doc_vec: document vector
        """
        return get_word_and_doc_vector(dtm)

    def fit(self, texts: List[List[str]], epochs: int = 200, first_chunk: bool = True,
            chunk_end: int = None, memory_start: int = 0, workers=1) -> None:
        """
        LDA Gibbs Sampler that uses a C-implementation under the hood
        Args:
            texts: list of tokenized texts
            epochs: number of epochs to train the models for
            first_chunk: when using a RollingLDA, is this the first time chunk? Can be ignored, if no RollingLDA is used.
            chunk_end: end index of the period to train the model for when using a RollingLDA. Can be ignored, if no RollingLDA is used.
            memory_start: index to start loading the assignment vectors for a RollingLDA from. Can be ignored, if no RollingLDA is used.
            workers: number of workers to use for parallelization
        Returns:
            None
        """
        if not isinstance(texts, list):
            if isinstance(texts, str):
                raise TypeError("texts must be a list of lists of strings!")
            try:
                texts = list(texts)
                if len(texts) <= 1:
                    raise ValueError
            except ValueError:
                raise TypeError("texts must be a list of lists of strings!")
        if not isinstance(texts[0], list):
            texts = [list(x) for x in texts]
        if not isinstance(texts[0][0], str):
            texts = [[str(x) for x in text] for text in texts]
            warnings.warn("The given texts are not input as strings.")
        if not isinstance(epochs, int):
            try:
                if epochs != int(epochs):
                    raise ValueError
                epochs = int(epochs)
            except ValueError:
                raise TypeError("epochs must be an integer!")
        if epochs < 1:
            raise ValueError("epochs must be a natural number greater than 0")
        if not isinstance(first_chunk, bool):
            raise TypeError("first_chunk must be a boolean!")
        if chunk_end is not None and not isinstance(chunk_end, int):
            try:
                if chunk_end != int(chunk_end):
                    raise ValueError
                chunk_end = int(chunk_end)
            except ValueError:
                raise TypeError("chunk_end must be an integer!")
        if chunk_end is not None and chunk_end < 0:
            raise ValueError("chunk_end must be a natural number greater than 0")
        if not isinstance(memory_start, int):
            try:
                if memory_start != int(memory_start):
                    raise ValueError
                memory_start = int(memory_start)
            except ValueError:
                raise TypeError("memory_start must be an integer!")
        if memory_start < 0:
            raise ValueError("memory_start must be a natural number greater than 0")
        if not isinstance(workers, int):
            try:
                if workers != int(workers):
                    raise ValueError
                workers = int(workers)
            except ValueError:
                raise TypeError("workers must be an integer!")
        if chunk_end is None:
            chunk_end = len(texts)
        texts = texts[self._last_end[-1]:chunk_end]
        old_dtm_size = self._dtm.shape[0] if self._dtm is not None else 0
        self._create_dtm(texts)

        word_vec, doc_vec = self._get_word_and_doc_vector(self._dtm[old_dtm_size:chunk_end+old_dtm_size, :])
        if self._verbose > 1:
            print("Word_vec length: ", len(word_vec))
        corpus_size = self._dtm.shape[0] - old_dtm_size
        vocab_size = len(self._vocab)

        if memory_start > 0:
            memory_start = sum([memory_start > x for x in self._deleted_indices])
            memory_start = sum(self._len_of_docs[:memory_start])
        if first_chunk:
            memory_word_topic_matrix = np.zeros((vocab_size, self._K), dtype=np.uint64)
        else:
            memory_word_vec = self._word_vec[memory_start:]
            memory_assignments = self._assignments[memory_start:]
            memory_word_topic_matrix = self.get_word_topic_matrix(memory_word_vec, memory_assignments)
            if self._verbose > 1:
                print(f"Loaded a memory of {len(memory_assignments)} assignments")

        document_topic_matrix = np.zeros((corpus_size, self._K), dtype=np.uint64)
        v_sum = np.sum(memory_word_topic_matrix, axis=0)

        self._calculate_prototype(word_vec, doc_vec, memory_word_topic_matrix,
                                  document_topic_matrix, v_sum, workers, epochs, old_dtm_size)
        self._is_trained = True
        self._doc_vec = np.concatenate((self._doc_vec, doc_vec), axis=0) if self._doc_vec is not None else doc_vec
        doc_id, tmp_counts = np.unique(self._doc_vec, return_counts=True)
        self._len_of_docs = np.zeros(np.array(self._doc_vec.max() + 1, dtype=np.uint64), dtype=np.uint32)
        self._len_of_docs[doc_id] = tmp_counts
        self._last_end.append(self._dtm.shape[0])

    def _calculate_prototype(self, word_vec: np.ndarray, doc_vec: np.ndarray, word_topic_matrix: np.ndarray,
                             document_topic_matrix: np.ndarray, v_sum: np.ndarray, workers: int, epochs: int = 200,
                             period_start: int = 0):
        """
        Calculates the LDAPrototype for the texts given in fit().
        Args:
            word_vec: vector of word ids in the vocabulary for each word
            doc_vec: vector of document ids for each word
            word_topic_matrix: word-topic matrix
            document_topic_matrix: document-topic matrix
            v_sum: sum of word-topic matrix
            workers: number of workers to use for parallelization
            epochs: number of epochs to train the models for
            period_start: index to start loading the assignment vectors for a RollingLDA from. Can be ignored, if no RollingLDA is used.
        """
        if not isinstance(word_vec, np.ndarray):
            raise TypeError("word_vec must be a numpy array!")
        if not isinstance(doc_vec, np.ndarray):
            raise TypeError("doc_vec must be a numpy array!")
        if not isinstance(word_topic_matrix, np.ndarray):
            raise TypeError("word_topic_matrix must be a numpy array!")
        if not isinstance(document_topic_matrix, np.ndarray):
            raise TypeError("document_topic_matrix must be a numpy array!")
        if not isinstance(v_sum, np.ndarray):
            raise TypeError("v_sum must be a numpy array!")
        if not isinstance(workers, int):
            raise TypeError("workers must be an integer!")
        if not isinstance(epochs, int):
            raise TypeError("epochs must be an integer!")
        if not isinstance(period_start, int):
            raise TypeError("period_start must be an integer!")
        if period_start < 0:
            raise ValueError("period_start must be a natural number greater than 0")

        all_results = []
        if workers > 1:
            assignments = np.random.randint(0, self._K, len(word_vec), dtype=np.uint32)
            with Pool(workers) as pool:
                all_results = pool.starmap(self._sample_gibbs, [(word_vec.copy(), assignments.copy(), doc_vec.copy(),
                                                                 word_topic_matrix.copy(), document_topic_matrix.copy(),
                                                                 v_sum.copy(), epochs) for _ in range(self._prototype)])
            all_results = list(all_results)
        else:
            iterator = range(self._prototype)
            if self._verbose > 0:
                iterator = tqdm(iterator)
            for sample in iterator:
                if self._verbose > 0:
                    iterator.set_description(f"Prototype {sample + 1}/{self._prototype}")
                assignments = np.random.randint(0, self._K, len(word_vec), dtype=np.uint32)
                res = self._sample_gibbs(word_vec.copy(), assignments.copy(), doc_vec.copy(), word_topic_matrix.copy(), document_topic_matrix.copy(),
                                         v_sum.copy(), epochs)
                all_results.append(res)

        if self._prototype > 1:
            current_word_topic_matrices = [self.get_word_topic_matrix(word_vec, x[0]) for x in all_results]
            sclop = TopicClusters(current_word_topic_matrices, self._measure, self._threshold, self._K).select_prototype()
            if sclop is None:
                new_assignments = np.random.randint(0, self._K, len(word_vec), dtype=np.uint32)
                prototype = [new_assignments.copy(), self.get_word_topic_matrix(word_vec, new_assignments), self.get_document_topic_matrix(doc_vec, new_assignments)]
            else:
                prototype = all_results[np.argmax(sclop)]
        else:
            prototype = all_results[0]

        doc_vec += period_start
        self._assignments = np.concatenate((self._assignments, prototype[0]), axis=0) if self._assignments is not None else prototype[0]
        self._word_vec = np.concatenate((self._word_vec, word_vec), axis=0) if self._word_vec is not None else word_vec
        self._document_topic_matrix.append(prototype[2])

    def _sample_gibbs(self, word_vec: np.ndarray, assignment_vec: np.ndarray, doc_vec: np.ndarray, word_topic_matrix: np.ndarray,
                      document_topic_matrix: np.ndarray, v_sum: np.ndarray, epochs: int = 200) -> (
            Tuple)[np.ndarray, np.ndarray, np.ndarray]:
        """
        LDA Gibbs Sampler that uses a C-implementation under the hood and edits the inputs in place
        Args:
            word_vec: word vector
            assignment_vec: assignment vector
            doc_vec: document vector
            word_topic_matrix: word-topic matrix
            document_topic_matrix: document-topic matrix
            v_sum: sum of word-topic matrix
            epochs: number of epochs to train the models for
        Returns:
            assignment_vec: assignment vector
            word_topic_matrix: word-topic matrix
            document_topic_matrix: document-topic matrix
        """
        if not isinstance(word_vec, np.ndarray):
            raise TypeError("word_vec must be a numpy array!")
        if not isinstance(doc_vec, np.ndarray):
            raise TypeError("doc_vec must be a numpy array!")
        if not isinstance(word_topic_matrix, np.ndarray):
            raise TypeError("word_topic_matrix must be a numpy array!")
        if not isinstance(document_topic_matrix, np.ndarray):
            raise TypeError("document_topic_matrix must be a numpy array!")
        if not isinstance(v_sum, np.ndarray):
            raise TypeError("v_sum must be a numpy array!")
        if not isinstance(epochs, int):
            raise TypeError("epochs must be an integer!")
        vanilla_gibbs_func(word_vec, assignment_vec, doc_vec, word_topic_matrix, document_topic_matrix, v_sum,
                           np.array(self._alpha.copy()), np.array(self._gamma.copy()), np.zeros((self._K,)), self._K, epochs, 0)

        if self._max_assign:
            final_assignment_func(word_vec, assignment_vec, doc_vec, word_topic_matrix, document_topic_matrix, v_sum, self._alpha.copy(), self._gamma.copy(),
                                  self._K, epochs, 0)
        return assignment_vec, word_topic_matrix, document_topic_matrix

    def get_word_topic_matrix(self, word_vec: np.ndarray = None, assignment_vec: np.ndarray = None) -> np.ndarray:
        """
        Create a word-topic matrix from a word vector and an assignment vector
        Args:
            word_vec: word vector
            assignment_vec: assignment vector
        Returns:
            word_topic_matrix: word-topic matrix
        """
        if word_vec is None:
            word_vec = self._word_vec
        elif not isinstance(word_vec, np.ndarray):
            try:
                word_vec = np.array(word_vec, dtype=np.uint64)
            except ValueError:
                raise TypeError("word_vec must be a numpy array!")
        if assignment_vec is None:
            assignment_vec = self._assignments
        elif not isinstance(assignment_vec, np.ndarray):
            try:
                assignment_vec = np.array(assignment_vec, dtype=np.uint32)
            except ValueError:
                raise TypeError("assignment_vec must be a numpy array!")
        if word_vec.shape != assignment_vec.shape:
            raise ValueError("word_vec and assignment_vec must have the same shape!")
        word_topic_matrix = np.zeros((len(self._vocab), self._K), dtype=np.uint64)
        load_wk_mat_func(word_vec, assignment_vec, word_topic_matrix, self._K)
        return word_topic_matrix.copy()

    def get_document_topic_matrix(self, doc_vec: np.ndarray, assignment_vec: np.ndarray) -> np.ndarray:
        """
        Create a word-topic matrix from a word vector and an assignment vector
        Args:
            doc_vec: document vector
            assignment_vec: assignment vector
        Returns:
            document_topic_matrix: document-topic matrix
        """
        if not isinstance(doc_vec, np.ndarray):
            try:
                doc_vec = np.array(doc_vec, dtype=np.uint64)
            except ValueError:
                raise TypeError("doc_vec must be a numpy array!")
        if not isinstance(assignment_vec, np.ndarray):
            try:
                assignment_vec = np.array(assignment_vec, dtype=np.uint32)
            except ValueError:
                raise TypeError("assignment_vec must be a numpy array!")
        if doc_vec.shape != assignment_vec.shape:
            raise ValueError("doc_vec and assignment_vec must have the same length!")
        doc_vec = doc_vec - min(doc_vec)  # Getting a zero-based index
        document_topic_matrix = np.zeros((len(set(doc_vec)), self._K), dtype=np.uint64)
        load_dk_mat_func(doc_vec, assignment_vec, document_topic_matrix, self._K)
        return document_topic_matrix.copy()

    def top_words(self, number: int = 5, topic: int = None, importance: Union[bool, np.ndarray] = True,
                  word_topic_matrix: np.ndarray = None, return_as_data_frame: bool = True) -> Union[List[str], List[List[str]], pd.DataFrame]:
        """
        Get the top words for a topic
        Args:
            number: number of top words to return
            topic: topic to return top words for or None for all topics
            importance: bool: should the importance or absolute frequency be used to determine the top words
                        np.array: importance of words for each topic
            word_topic_matrix: word-topic matrix
            return_as_data_frame: should the result be returned as a pandas DataFrame (Default) or a list?
        Returns:
            top_words: top words
        """
        if not isinstance(number, int):
            try:
                if number == int(number):
                    number = int(number)
                else:
                    raise ValueError
            except ValueError:
                raise TypeError("number must be an integer!")
        if number < 1:
            raise ValueError("number must be a natural number greater than 0")
        if topic is not None and not isinstance(topic, int):
            try:
                if int(topic) != topic:
                    raise ValueError
                topic = int(topic)
            except ValueError:
                raise TypeError("topic must be an integer!")
        if topic is not None and (topic < 1 or topic > self._K):
            raise ValueError("topic must be a natural number between 1 and K")
        if not isinstance(importance, bool) and not isinstance(importance, np.ndarray):
                raise TypeError("importance must be a boolean!")
        if word_topic_matrix is not None and not isinstance(word_topic_matrix, np.ndarray):
            raise TypeError("word_topic_matrix must be a numpy array!")
        if word_topic_matrix is None:
            if self._is_trained:
                word_topic_matrix = self.get_word_topic_matrix(self._word_vec, self._assignments)
            else:
                raise AttributeError("word_topic_matrix must be provided if the model has not been trained yet!")
        if not isinstance(return_as_data_frame, bool):
            raise TypeError("return_as_data_frame must be a boolean!")
        if isinstance(importance, bool) and importance:
            importance = calculate_importance(word_topic_matrix) if importance else None
        if topic is None:
            top_words = [self.top_words(number, k, importance, word_topic_matrix) for k in range(1, self._K + 1)]
            if return_as_data_frame:
                top_words = pd.DataFrame(np.asarray(top_words).transpose(), columns=[f"Topic {k}" for k in range(len(top_words))])
            return top_words
        else:
            if isinstance(importance, np.ndarray):
                return [self._vocab[index] for index in np.argsort(-importance[topic - 1, :])[:number]]
            return [self._vocab[index] for index in np.argsort(-word_topic_matrix[topic - 1, :])[:number]]

    def wordclouds(self, topic: int = None, number: int = 50, path: str = "wordclouds.pdf", height: int = 600, width: int = 700,
                   show: bool = True, word_topic_matrix: np.ndarray = None) -> None:
        """
        Create a word cloud from a topic
        Args:
            topic: The topic to create a word cloud for or None for all topics
            number: number of top words to include in the word cloud
            path: path to a file to store the wordcloud files in
            width: width of the image
            height: height of the image
            show: should the image be shown
            word_topic_matrix: Optional. If you want the wordclouds to be based on a specific word-topic matrix, you can pass it here.
        Returns:
            None
        """
        if not isinstance(topic, int) and topic is not None:
            try:
                if int(topic) != topic:
                    raise ValueError
                topic = int(topic)
            except ValueError:
                raise TypeError("topic must be an integer!")
        if topic is not None and (topic < 1 or topic > self._K):
            raise ValueError("topic must be a natural number between 1 and K")
        if not isinstance(number, int):
            try:
                if number != int(number):
                    raise ValueError
                number = int(number)
            except ValueError:
                raise TypeError("number must be an integer!")
        if number < 1:
            raise ValueError("number must be a natural number greater than 0")
        if path is not None and not isinstance(path, str):
            raise TypeError("path must be a string or None!")
        if not isinstance(height, int):
            try:
                if height != int(height):
                    raise ValueError
                height = int(height)
            except ValueError:
                raise TypeError("height must be an integer!")
        if height < 1:
            raise ValueError("height must be a natural number greater than 0")
        if not isinstance(width, int):
            try:
                if width != int(width):
                    raise ValueError
                width = int(width)
            except ValueError:
                raise TypeError("width must be an integer!")
        if width < 1:
            raise ValueError("width must be a natural number greater than 0")
        if not isinstance(show, bool):
            raise TypeError("show must be a boolean!")
        if path is not None:
            matplotlib.use('Agg')
        if path is None and show == False:
            raise ValueError("'path' and 'show' cannot both be None")
        if word_topic_matrix is None:
            word_topic_matrix = self.get_word_topic_matrix(self._word_vec, self._assignments)
        importance = calculate_importance(word_topic_matrix)
        if topic is None:
            word_weights = [{self._vocab[index]: importance[k, index] for index in np.argsort(-importance[k, :])[:number]} for k in range(self._K)]
        else:
            word_weights = [{self._vocab[index]: importance[topic, index] for index in np.argsort(-importance[topic, :])[:number]}]
        figures = []
        for i, cloud in enumerate(word_weights):
            if sum(cloud.values()) == 0:
                warnings.warn(f"Topic {i + 1} has no words! This might happen if there are too few words in the corpus, compared to the number of topics!")
                continue
            wordcloud = WordCloud(width=width, height=height, background_color='white').generate_from_frequencies(cloud)
            fig = plt.figure(figsize=(15, 15 * height / width), dpi=150)
            plt.imshow(wordcloud, interpolation='bilinear')
            plt.axis('off')
            if topic is None:
                plt.title(f'Word Cloud {i + 1}')
            else:
                plt.title(f'Word Cloud {topic + 1}')
            figures.append(fig)
            if show:
                fig.show()
        if path is not None:
            with PdfPages(path) as pdf:
                for i, fig in enumerate(figures):
                    pdf.savefig(fig, bbox_inches='tight')
                    plt.close(fig)

    def get_assignment_vec(self) -> np.ndarray:
        """
        Returns:
            assignment vector
        """
        return self._assignments.copy()

    def get_word_vec(self) -> np.ndarray:
        """
        Returns:
            word vector
        """
        return self._word_vec.copy()

    def get_doc_vec(self) -> np.ndarray:
        """
        Returns:
            document vector
        """
        return self._doc_vec.copy()

    def get_vocab(self) -> List[str]:
        """
        Returns:
            vocabulary
        """
        return self._vocab.copy()

    def get_params(self) -> dict:
        """
        Returns:
            all parameters as a dictionary
        """
        return self.__dict__.copy()

    def set_params(self, params: dict) -> None:
        """
        Sets all parameters given py a dictionary to the parameters of the same name in the class
        Returns:
            None
        """
        if not isinstance(params, dict):
            raise TypeError("parameters must be a dictionary!")
        self.__dict__.update(params)

    def save(self, path: str) -> None:
        """
        Saves the model to a pickle file
        Returns:
            None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        pickle.dump(self, open(path, "wb"))

    def load(self, path: str) -> None:
        """
        Loads a model from a pickle file
        Returns:
            None
        """
        if not isinstance(path, str):
            raise TypeError("path must be a string!")
        loaded = pickle.load(open(path, "rb"))
        self.__dict__ = loaded.__dict__

    def shrink_model(self, new_start_index: int) -> None:
        """
        Shrinks the model to a new start index to save memory
        """
        # todo implement
        raise NotImplementedError

    def is_trained(self) -> bool:
        """
        Returns:
            Boolean if the model has been trained before
        """
        return self._is_trained



import os.path

import torch
from pathlib import Path
from transformers import BertTokenizer, BertModel
from transformers import logging as lg
from ttta.preprocessing.schemas import OxfordAPIResponse, WordSenseEmbedding
from ttta.preprocessing.settings import FileLoader
import logging
from typing import Union


class VectorEmbeddings:
    """
    This class is used to infer the vector embeddings of a word from a sentence.

    Methods
    -------
        infer_vector(doc:str, main_word:str)
            This method is used to infer the vector embeddings of a word from a sentence.
        _bert_case_preparation()
            This method is used to prepare the BERT model for the inference.
    """
    def __init__(
        self,
        pretrained_model_path: Union[str, Path] = None,
    ):
        self.model_path = pretrained_model_path
        if pretrained_model_path is not None:
            if not os.path.exists(pretrained_model_path):
                raise ValueError(
                    f'The path {pretrained_model_path} does not exist'
                )
            self.model_path = Path(pretrained_model_path)

        self._tokens = []
        self.model = None
        self.vocab = False
        self.lematizer = None

        lg.set_verbosity_error()
        self._bert_case_preparation()

    @property
    def tokens(self):
        return self._tokens

    def _bert_case_preparation(self) -> None:
        """
        This method is used to prepare the BERT model for the inference.
        """
        model_path = self.model_path if self.model_path is not None else 'bert-base-uncased'
        self.bert_tokenizer = BertTokenizer.from_pretrained(model_path)
        self.model = BertModel.from_pretrained(
            model_path,
            output_hidden_states = True,
        )
        self.model.eval()
        self.vocab = True

    def infer_vector(self, doc:str, main_word:str):
        """
        This method is used to infer the vector embeddings of a word from a sentence.
        Args:
            doc: Document to process
            main_word: Main work to extract the vector embeddings for.

        Returns: torch.Tensor

        """
        if not self.vocab:
            raise ValueError(
                f'The Embedding model {self.model.__class__.__name__} has not been initialized'
            )
        marked_text = "[CLS] " + doc + " [SEP]"
        tokens = self.bert_tokenizer.tokenize(marked_text)
        try:
            main_token_id = tokens.index(main_word.lower())
            idx = self.bert_tokenizer.convert_tokens_to_ids(tokens)
            segment_id = [1] * len(tokens)

            self.tokens_tensor = torch.tensor([idx])
            self.segments_tensors = torch.tensor([segment_id])

            with torch.no_grad():
                outputs = self.model(self.tokens_tensor, self.segments_tensors)
                hidden_states = outputs[2]

            return hidden_states[-2][0][main_token_id]

        except ValueError:
            raise ValueError(
                f'The word: "{main_word}" does not exist in the list of tokens: {tokens} from {doc}'
            )


class ExtractSenseEmbeddings:
    """
    Wrapper class for the Vector embeddings that is used to extract the embeddings for all the senses.

    Attributes
    ----------
        sense: dict
            The sense for which the embeddings are to be extracted.
        word: str
            The word for which the embeddings are to be extracted.
        vector_embeddings: VectorEmbeddings
            The class that is used to extract the embeddings.
        api_component: OxfordAPIResponse
            The class that is used to extract the senses.
        all_words: list
            The list of all the words for which the embeddings are to be extracted.

    Methods
    -------
        __call__(sense:dict, main_w:str)
            This method is used to initialize the object.
        _infer_sentence_vector()
            This method is used to infer the vector embeddings for all the sentences in the sense.
        infer_mean_vector()
            This method is used to infer the mean vector embeddings for all the sentences in the sense.
        create_sense_embeddings()
            This method is used to extract the embeddings for all the senses.

    """
    def __init__(
            self
        ):
        self.sense = None
        self.word = None
        self.vector_embeddings = VectorEmbeddings()
        self.api_component = OxfordAPIResponse()
        self.all_words = FileLoader.load_files(self.__class__.__name__)

    def __call__(self, sense:dict, given_word:str) -> 'ExtractSenseEmbeddings':
        """
        This method is used to initialize the object with the particular senses and the given word.
        Args:
            sense: The sense for which the embeddings are to be extracted.
            given_word: The word to track in each examples for all senses of the word.

        Returns: The self object initialized.

        """
        if not isinstance(sense, OxfordAPIResponse):
            raise ValueError(
                f'Expected type {OxfordAPIResponse.__class__} for the sense, but got type: {type(sense)}'
            )
        self.sense = sense
        self.word = given_word
        return self

    def _infer_sentence_embedding(self) -> torch.Tensor:
        """
        Infer the embeddings of the give_word in each example of the sense.
        Returns: torch.Tensor

        """
        for example in self.sense.examples:
            yield self.vector_embeddings.infer_vector(
                doc=example,
                main_word=self.word
            )

    def infer_mean_vector(self) -> WordSenseEmbedding:
        """
        Infer the mean vector embedding for the given word across all the examples in the senses.
        Returns: WordSenseEmbedding object.

        """
        all_token_embeddings =  torch.stack(list(self._infer_sentence_embedding()))
        return WordSenseEmbedding(
            id=self.sense.id,
            definition=self.sense.definition,
            embedding=torch.mean(all_token_embeddings, dim=0).tolist(),
        )

    def create_sense_embeddings(self) -> list:
        """
        Extract the averaged vector embedding for a list of polysemous words across
        a wide range of contexts.
        Returns:
            list: A list of dictionaries containing the word and the sense embeddings.

        """
        all_embeddings = []
        word_embedding = {}
        for word_obj in self.all_words:
            logging.info(f'{"-"*40} Embedding the word {word_obj.word} {"-"*40} ')
            word_embedding['word'] = word_obj.word
            word_embedding['senses'] = [self(sens, word_obj.word).infer_mean_vector() for sens in word_obj.senses]
            all_embeddings += [word_embedding.copy()]

        return all_embeddings


import numpy as np
from typing import Union, List, Tuple, Set
from scipy.cluster.hierarchy import linkage
import warnings
from math import floor
import scipy
import pandas as pd


class TopicClusters:
    def __init__(self, word_topic_matrices: List[np.ndarray] = None, measure: Union[str, callable] = "jaccard",
                 topic_threshold: List[Union[int, float]] = None, K: int = 30) -> None:
        """
        Implements the TopicClustering class. This class is used to cluster topics from topic models based on a given measure.
        It returns a score for each individual model pair, which can be used to select a prototype from the list of models.
        Apart from that, it can also give a summary as to which models have which topics in common.
        Args:
            word_topic_matrices: list of word-topic matrices
            measure: measure to use for prototype selection
            threshold_dict: dictionary containing the indices of the words that occur more than the threshold in a topic
        """
        if not isinstance(word_topic_matrices, list):
            raise TypeError("word_topic_matrices must be a list!")
        if not isinstance(word_topic_matrices[0], np.ndarray):
            raise TypeError("word_topic_matrices must be a list of numpy arrays!")
        if len(word_topic_matrices) < 2:
            raise ValueError("word_topic_matrices must contain at least two word-topic matrices!")
        self._word_topic_matrices = word_topic_matrices
        if not isinstance(measure, str) and not callable(measure):
            raise TypeError("measure must be a string or a callable!")
        self._measure = measure
        if topic_threshold is None:
            topic_threshold = [5, 0.002]
        elif not isinstance(topic_threshold, list):
            raise TypeError("topic_threshold must be a list of an integer and a float!")
        elif topic_threshold[0] < 0 or topic_threshold[1] < 0 or topic_threshold[1] > 1:
            raise ValueError("topic_threshold must contain a positive integer and a float between 0 and 1!")
        self._threshold = topic_threshold
        if not isinstance(K, int):
            try:
                if K == int(K):
                    K = int(K)
                else:
                    raise ValueError
            except:
                raise TypeError("K must be an integer!")
        elif K < 2:
            raise ValueError("K must be a natural number greater than 1")
        self._K = K
        if word_topic_matrices[0].shape[1] != K:
            raise ValueError("The number of word-topic matrices must be a multiple of K!")
        self._number_of_models = len(word_topic_matrices)
        self._vocab_length = word_topic_matrices[0].shape[0]
        self.sclops = None
        self.matched_topics = None

    def select_prototype(self) -> Union[np.ndarray, None]:
        """
        Selects the prototype from a list of word-topic matrices.
        Returns:
            index of the prototype in the list of word-topic matrices
        """
        if self.sclops is None:
            self._compare_topics(self._word_topic_matrices, self._measure)
        return self.sclops if self.sclops is None else np.argmax([np.mean(x) for x in self.sclops])

    def get_matched_topics(self) -> Union[pd.DataFrame, None]:
        """
        Returns a dataframe containing all matched topics for a list of word-topic matrices
        Returns:
            dataframe containing all matched topics for a list of word-topic matrices
        """
        if self.matched_topics is None:
            self._compare_topics(self._word_topic_matrices, self._measure)
        return self.matched_topics

    def _compare_topics(self, word_topic_matrices: List[np.ndarray], measure: Union[str, callable]) -> None:
        """
        Compares the topics of the list of word-topic matrices, given a measure. Can be used to select a prototype
        or to give a summary of common topics.
        Args:
            word_topic_matrices: list of word-topic matrices
            measure: measure to use for prototype selection
        Returns:
            index of the prototype in the list of word-topic matrices
        """
        word_topic_matrices = np.vstack([x.transpose() for x in word_topic_matrices])
        threshold_dict, broken_indices = self._check_topic_thresholds(word_topic_matrices) if measure == "jaccard" else (None, set())
        if len(broken_indices) == self._number_of_models:
            warnings.warn(f"All models are broken. This can happen when there are too few documents to "
                          f"properly train the LDA or the words occur not often enough to satisfy the topic_threshold. Please consider"
                          f"lowering the values in topic_threshold. The assignments will be randomly sampled.")
            return None
        elif len(broken_indices) > 0:
            warnings.warn(f"Encountered {len(broken_indices)} broken models. This can happen when there are too few "
                          f"documents to properly train the LDA or the 'topic_threshold'-parameter has been set too "
                          f"restrictively. Please consider lowering the values in topic_threshold. "
                          f"The broken models will be ignored for the prototype selection.")
        sclops = [[] for _ in range(self._number_of_models)]
        matched_topics = None
        for i in range(self._number_of_models - 1):
            for j in range(i + 1, self._number_of_models):
                if i in broken_indices or j in broken_indices:
                    sclops[i].append(-1)
                    sclops[j].append(-1)
                    continue
                temp_matched_topics, sclop = self._cluster_topics(i, j, measure, threshold_dict)
                if matched_topics is None:
                    matched_topics = temp_matched_topics
                else:
                    matched_topics = pd.concat([matched_topics, temp_matched_topics])
                sclops[i].append(sclop)
                sclops[j].append(sclop)
        self.sclops = sclops
        self.matched_topics = matched_topics

    def _check_topic_thresholds(self, word_topic_matrices: np.ndarray) -> Tuple[dict, Set[int]]:
        """
        Filters out words that occur less than the threshold in a topic for the prototype selection.
        Args:
            word_topic_matrices: list of word-topic matrices
        Returns:
            threshold_dict: dictionary containing the indices of the words that occur more than the threshold in a topic
            broken_models: Index of all broken models that are to be ignored, because they store all words into one topic
                           (happens rarely when there are too few documents).
        """
        row_sums = np.sum(word_topic_matrices, axis=1)
        row_sums = np.where(row_sums < 1, 1, row_sums)
        vk_thresholds = np.argwhere((word_topic_matrices >= self._threshold[0]) & (word_topic_matrices / row_sums[:, None] >= self._threshold[1]))
        threshold_dict = {x: list(set(vk_thresholds[np.argwhere(vk_thresholds[:, 0] == x).flatten(), 1])) for x in range(self._number_of_models * self._K)}
        zero_indices = np.array([key for key, value in threshold_dict.items() if len(value) == 0])
        broken_models = set([floor(index / self._K) for index in zero_indices])
        return threshold_dict, broken_models

    def _cluster_topics(self, i: int, j: int, measure: Union[str, callable], threshold_dict: dict) -> Tuple[pd.DataFrame, float]:
        """
        Clusters topics given their word-topic matrices and returns the portion of matched pairs stemming from different models
        Args:
            i: index of the first topic
            j: index of the second topic
            measure: measure to use for prototype selection
            threshold_dict: dictionary containing the indices of the words that occur more than the threshold in a topic
        Returns:
            portion of matched pairs stemming from different models
        """
        if not isinstance(i, int):
            raise TypeError("i must be an integer!")
        if not isinstance(j, int):
            raise TypeError("j must be an integer!")
        if not isinstance(threshold_dict, dict):
            raise TypeError("threshold_dict must be a dictionary!")
        if not isinstance(measure, str) and not callable(measure):
            raise TypeError("measure must be a string or a callable!")
        temp_mat = np.zeros((2 * self._K, self._vocab_length), dtype=int)
        if measure == "jaccard":
            for row in range(self._K):
                temp_mat[row, threshold_dict[row + i * self._K]] = 1
                temp_mat[row + self._K, threshold_dict[row + j * self._K]] = 1

        temp_distances = scipy.spatial.distance.pdist(temp_mat, metric=measure)
        linkage_result = linkage(temp_distances, method='complete')
        linkage_result = pd.DataFrame(linkage_result[:, :2], columns=['topic_one', 'topic_two']).astype(int)
        linkage_result["model_one"] = i
        linkage_result["model_two"] = j
        linkage_result = linkage_result[linkage_result['topic_one'] < 2 * self._K]
        linkage_result = linkage_result[linkage_result['topic_two'] < 2 * self._K]
        matched_topics = linkage_result[[(linkage_result['topic_one'][i] < self._K <= linkage_result['topic_two'][i]) or
                                                       (linkage_result['topic_two'][i] < self._K <= linkage_result['topic_one'][i])
                                                       for i, _ in linkage_result.iterrows()]]
        number_of_matched_topics = len(matched_topics)
        return matched_topics, 1 - (self._K - number_of_matched_topics) / (2 * self._K)